import numpy as np
import pint
import pytest

from eradiate import unit_registry as ureg
from eradiate._factory import Factory
from eradiate.scenes.core import (
    BoundingBox,
    get_factory,
)


def test_bbox():
    # Instantiation with correctly ordered unitless values works
    bbox = BoundingBox([0, 0, 0], [1, 1, 1])
    assert bbox.units == ureg.m

    # Instantiation with correctly ordered unit-attached values works
    bbox = BoundingBox([0, 0, 0] * ureg.m, [1, 1, 1] * ureg.m)
    assert bbox.min.units == ureg.m

    # Extents are correctly computed
    assert np.allclose([1, 2, 3] * ureg.m, BoundingBox([0, 0, 0], [1, 2, 3]).extents)

    # Extent shapes must be compatible
    with pytest.raises(ValueError):
        BoundingBox([0, 0], [1, 1, 1])
    with pytest.raises(ValueError):
        BoundingBox([0, 0, 0], [[1, 1, 1]])

    # Unit mismatch raises
    with pytest.raises(pint.DimensionalityError):
        BoundingBox([0, 0] * ureg.m, [1, 1] * ureg.s)


def test_bbox_convert():
    bbox_ref = BoundingBox([0, 0, 0], [1, 1, 1])

    bbox_convert = BoundingBox.convert([[0, 0, 0], [1, 1, 1]])
    assert np.allclose(bbox_convert.min, bbox_ref.min)
    assert np.allclose(bbox_convert.max, bbox_ref.max)

    bbox_convert = BoundingBox.convert(np.array([[0, 0, 0], [1, 1, 1]]))
    assert np.allclose(bbox_convert.min, bbox_ref.min)
    assert np.allclose(bbox_convert.max, bbox_ref.max)

    bbox_ref = BoundingBox([0, 0, 0] * ureg.m, [1, 1, 1] * ureg.m)
    bbox_convert = BoundingBox.convert([[0, 0, 0], [1, 1, 1]] * ureg.m)
    assert np.allclose(bbox_convert.min, bbox_ref.min)
    assert np.allclose(bbox_convert.max, bbox_ref.max)


def test_bbox_contains():
    bbox = BoundingBox([0, 0, 0], [1, 1, 1])

    # Works with a single point
    assert bbox.contains([0.5, 0.5, 0.5])
    assert not bbox.contains([0.5, 0.5, -0.5])
    assert not bbox.contains([0.5, 0.5, 0.5] * ureg.km)

    # Works with multiple points
    assert np.all(bbox.contains([[0.5, 0.5, 0.5], [0.5, -0.5, 0.5]]) == [True, False])


def test_get_factory():
    """
    Check that all declared factories can be retrieved.
    """
    for element_type in [
        "atmosphere",
        "biosphere",
        "bsdf",
        "illumination",
        "integrator",
        "measure",
        "phase",
        "shape",
        "spectrum",
        "surface",
    ]:
        assert isinstance(get_factory(element_type), Factory)



import drjit as dr
import mitsuba as mi

from eradiate.kernel import map_cube, map_unit_cube


def test_map_unit_cube(mode_mono):
    """
    Returns a transformation that maps old cube vertices to new cube vertices.
    """
    trafo = map_unit_cube(1, 2, 3, 4, 5, 6)

    vertices = [
        [0, 0, 0],
        [1, 0, 0],
        [0, 1, 0],
        [1, 1, 0],
        [0, 0, 1],
        [1, 0, 1],
        [0, 1, 1],
        [1, 1, 1],
    ]

    new_vertices = [
        [1, 3, 5],
        [2, 3, 5],
        [1, 4, 5],
        [2, 4, 5],
        [1, 3, 6],
        [2, 3, 6],
        [1, 4, 6],
        [2, 4, 6],
    ]

    for vertex, new_vertex in zip(vertices, new_vertices):
        assert dr.allclose(
            trafo.transform_affine(mi.Point3f(vertex)),
            new_vertex,
        )


def test_map_cube(mode_mono):
    """
    Returns a transformation that maps old cube vertices to new cube vertices.
    """

    trafo = map_cube(1, 2, 3, 4, 5, 6)

    vertices = [
        [-1, -1, -1],
        [1, -1, -1],
        [-1, 1, -1],
        [1, 1, -1],
        [-1, -1, 1],
        [1, -1, 1],
        [-1, 1, 1],
        [1, 1, 1],
    ]

    new_vertices = [
        [1, 3, 5],
        [2, 3, 5],
        [1, 4, 5],
        [2, 4, 5],
        [1, 3, 6],
        [2, 3, 6],
        [1, 4, 6],
        [2, 4, 6],
    ]

    for vertex, new_vertex in zip(vertices, new_vertices):
        assert dr.allclose(
            trafo.transform_affine(mi.Point3f(vertex)),
            new_vertex,
        )



import numpy as np
import pytest

from eradiate import unit_registry as ureg
from eradiate.radprops import ZGrid


@pytest.mark.parametrize(
    "levels",
    [
        np.linspace(0, 100, 11) * ureg.km,
        np.linspace(0, 100000, 11),
    ],
    ids=["pint", "unitless"],
)
def test_zgrid_all(levels):
    zgrid = ZGrid(levels=levels)

    np.testing.assert_array_equal(
        zgrid.levels.m_as("km"), [0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100]
    )
    assert zgrid.n_levels == 11

    np.testing.assert_array_equal(
        zgrid.layers.m_as("km"), [5, 15, 25, 35, 45, 55, 65, 75, 85, 95]
    )
    assert zgrid.n_layers == 10

    assert zgrid.layer_height == 10 * ureg.km
    assert zgrid.total_height == 100 * ureg.km


def test_zgrid_fails():
    with pytest.raises(ValueError, match="levels must be regularly spaced"):
        ZGrid([0, 1, 3])



from __future__ import annotations

from ..scenes.atmosphere import Atmosphere
from ..scenes.bsdfs import BSDF, bsdf_factory
from ..scenes.measure import (
    DistantMeasure,
    Measure,
    MultiRadiancemeterMeasure,
    RadiancemeterMeasure,
)
from ..scenes.shapes import RectangleShape
from ..scenes.surface import BasicSurface, Surface, surface_factory


def measure_inside_atmosphere(atmosphere: Atmosphere, measure: Measure) -> bool:
    """
    Evaluate whether a sensor is placed within an atmosphere.

    Raises a ValueError if called with a :class:`.MultiRadiancemeterMeasure`
    with origins both inside and outside the atmosphere.
    """
    if atmosphere is None:
        return False

    shape = atmosphere.shape

    if isinstance(measure, MultiRadiancemeterMeasure):
        inside = shape.contains(measure.origins)

        if all(inside):
            return True

        elif not any(inside):
            return False

        else:
            raise ValueError(
                "Inconsistent placement of MultiRadiancemeterMeasure origins. "
                "Origins must lie either all inside or all outside of the "
                "atmosphere."
            )

    elif isinstance(measure, DistantMeasure):
        # Note: This will break if the user makes something weird such as using
        # a large offset value which would put some origins outside and others
        # inside the atmosphere shape
        return not measure.is_distant()

    elif isinstance(measure, RadiancemeterMeasure):
        return shape.contains(measure.origin)

    else:
        # Note: This will likely break if a new measure type is added
        return shape.contains(measure.origin)


def surface_converter(value: dict | Surface | BSDF) -> Surface:
    """
    Attempt to convert the surface specification into a surface type.

    Surfaces can be defined purely by their BSDF, in which case Eradiate will
    define a rectangular surface and attach that BSDF to it.
    """
    if isinstance(value, dict):
        try:
            # First, attempt conversion to BSDF
            value = bsdf_factory.convert(value)
        except ValueError:  # Type ID could not be found in BSDF factory registry
            # Attempt conversion to Surface
            return surface_factory.convert(value)

    # If we make it to this point, it means that dict conversion has been
    # performed with success
    if isinstance(value, BSDF):
        return BasicSurface(
            shape=RectangleShape(),
            bsdf=value,
        )

    return value



import mitsuba as mi
import numpy as np
import pytest

from eradiate import KernelContext
from eradiate import unit_registry as ureg
from eradiate.scenes.core import traverse
from eradiate.scenes.geometry import SceneGeometry
from eradiate.scenes.phase import BlendPhaseFunction
from eradiate.spectral.index import SpectralIndex
from eradiate.test_tools.types import check_scene_element


def mi_to_numpy(d):
    result = {}

    for k, v in d.items():
        if isinstance(v, mi.VolumeGrid):
            result[k] = np.array(v)
        elif isinstance(v, mi.Transform4f):
            result[k] = np.array(v.matrix)
        else:
            result[k] = v

    return result


def assert_cmp_dict(value, expected):
    """
    Helper function to compare two dictionaries.
    """
    # Keys should be identical
    assert set(value.keys()) == set(expected.keys())

    for key in value.keys():
        first_element = value[key]
        second_element = expected[key]

        if isinstance(first_element, np.ndarray) or isinstance(
            second_element, np.ndarray
        ):
            # Handle special case of NumPy arrays
            np.testing.assert_allclose(
                first_element, second_element, err_msg=f"Failing key: '{key}'"
            )

        else:
            # Other types are compared for proper equality
            np.testing.assert_equal(
                first_element, second_element, err_msg=f"failing key: '{key}'"
            )

        assert True


def test_blend_phase_construct_basic():
    """
    Test instantiation with simple weight values.
    """
    # Constructing with reasonable settings succeeds
    phase = BlendPhaseFunction(
        components=[{"type": "isotropic"}, {"type": "rayleigh"}, {"type": "hg"}],
        weights=[0.25, 0.25, 0.5],
    )
    assert phase.weights.shape == (3,)  # Scalar weights are stored as a 1D array

    # Improper number of components raises
    with pytest.raises(ValueError):
        BlendPhaseFunction(
            components=[{"type": "isotropic"}],
            weights=[1],
        )

    # Improper weight array shape raises
    with pytest.raises(ValueError):
        BlendPhaseFunction(
            components=[{"type": "isotropic"}, {"type": "rayleigh"}, {"type": "hg"}],
            weights=[0.25, 0.25],
        )
    with pytest.raises(ValueError):
        BlendPhaseFunction(
            components=[{"type": "isotropic"}, {"type": "rayleigh"}, {"type": "hg"}],
            weights=[[0.25, 0.25, 0.5]],
        )


def test_blend_phase_construct_array(modes_all_double):
    """
    Test instantiation and kernel dict generation with an array of weights.
    """
    # Constructing using 1D arrays for weights succeeds
    phase = BlendPhaseFunction(
        components=[
            {"type": "isotropic"},
            {"type": "isotropic"},
            {"type": "isotropic"},
        ],
        weights=[np.ones((10,)), np.ones((10,)), np.ones((10,))],
        geometry="plane_parallel",
    )
    assert phase.weights.shape == (3, 10)  # Array weights are stored as a 2D array


@pytest.mark.parametrize(
    "weights, expected",
    [
        (
            [[0, 1, 2], [1, 0, 2]],
            [[1, 0, 0.5]],
        ),
        (
            [[0, 1, 2], [1, 0, 2], [0.5, 0.5, 2]],
            [[1, 1 / 3, 2 / 3], [1 / 3, 1, 0.5]],
        ),
    ],
    ids=[
        "array_2cmp",
        "array_3cmp",
    ],
)
def test_blend_phase_weights(mode_mono, weights, expected):
    phase = BlendPhaseFunction(
        components=[{"type": "isotropic"}] * len(weights),
        weights=weights,
        geometry="plane_parallel",
    )
    np.testing.assert_allclose(
        phase.eval_conditional_weights(SpectralIndex.new()), expected
    )


@pytest.mark.parametrize(
    "geometry",
    [
        {
            "type": "plane_parallel",
            "width": 1.0 * ureg.m,
            "ground_altitude": 0.0 * ureg.m,
            "toa_altitude": 1.0 * ureg.m,
        },
        {
            "type": "spherical_shell",
            "planet_radius": 1.0 * ureg.m,
            "ground_altitude": 0.0 * ureg.m,
            "toa_altitude": 1.0 * ureg.m,
        },
    ],
    ids=["plane_parallel", "spherical_shell"],
)
def test_blend_phase_geometry(mode_mono, geometry):
    """
    Test geometry and grid volume transform.
    """
    # Defining no geometry is accepted; in that case, no transform is applied
    # to the volume data
    phase = BlendPhaseFunction(
        components=[{"type": "isotropic"}, {"type": "rayleigh"}, {"type": "hg"}],
        weights=[0.25, 0.25, 0.5],
    )
    assert phase.geometry is None
    template = traverse(phase)[0]
    assert "weight.to_world" not in template

    # Appropriate geometry setup works
    geometry = SceneGeometry.convert(geometry)
    expected = np.array(geometry.atmosphere_volume_to_world.matrix)
    phase = BlendPhaseFunction(
        components=[{"type": "isotropic"}, {"type": "rayleigh"}, {"type": "hg"}],
        weights=[0.25, 0.25, 0.5],
        geometry=geometry,
    )
    template = traverse(phase)[0]
    to_world = np.array(template["weight.to_world"].matrix)
    np.testing.assert_allclose(to_world, expected)
    check_scene_element(phase, mi.PhaseFunction)

    # Nested BlendPhaseFunction objects must have the same geometry
    phase = BlendPhaseFunction(
        components=[
            {"type": "isotropic"},
            {
                "type": "blend_phase",
                "components": [{"type": "rayleigh"}, {"type": "hg"}],
                "weights": [0.25, 0.75],
            },
        ],
        weights=[0.25, 0.75],
        geometry=geometry,
    )

    for comp in phase.components:
        if isinstance(comp, BlendPhaseFunction):
            template = traverse(comp)[0]
            to_world = np.array(template["weight.to_world"].matrix)
            np.testing.assert_allclose(to_world, expected)

    check_scene_element(phase, mi.PhaseFunction)


@pytest.mark.parametrize(
    "kwargs",
    [
        {"weights": [0.0, 0.0]},
        {"weights": [1.0, 0.0]},
        {"weights": [0.0, 1.0]},
        {"weights": [0.5, 0.5]},
        {"weights": [0.3, 0.7]},
        {
            "weights": [
                np.array([1.0, 0.7, 0.5, 0.3, 0.0]),
                np.array([0.0, 0.3, 0.5, 0.7, 1.0]),
            ],
            "geometry": "plane_parallel",
        },
    ],
    ids=[
        "scalar0",
        "scalar1",
        "scalar2",
        "scalar3",
        "scalar4",
        "array",
    ],
)
def test_blend_phase_kernel_dict_2_components(mode_mono, kwargs):
    """
    Blendphase with 2 components produces correct kernel dict and can be loaded.
    """
    phase = BlendPhaseFunction(
        components=[{"type": "isotropic"}, {"type": "rayleigh"}],
        **kwargs,
    )
    check_scene_element(phase, mi.PhaseFunction)

    template, params = traverse(phase)
    ctx = KernelContext()
    kernel_dict = mi_to_numpy(template.render(ctx, nested=False))

    # Check that the kernel dict is correct
    expected = {
        "type": "blendphase",
        "phase_0.type": "isotropic",
        "phase_1.type": "rayleigh",
        "weight.type": "gridvolume",
        "weight.grid": np.reshape(phase.eval_conditional_weights(ctx.si), (-1, 1, 1)),
    }
    if "geometry" in kwargs:
        geometry = SceneGeometry.convert(kwargs["geometry"])
        expected["weight.to_world"] = np.array(
            geometry.atmosphere_volume_to_world.matrix
        )
    assert_cmp_dict(kernel_dict, expected)

    # Check that the parameter map is correct
    assert set(params.keys()) == {"weight.data"}


@pytest.mark.parametrize(
    "kwargs, expected_mi_weights",
    [
        ({"weights": [0.0, 0.0, 0.0]}, [0.0, 0.0]),
        ({"weights": [1.0, 0.0, 0.0]}, [0.0, 0.0]),
        ({"weights": [0.0, 1.0, 0.0]}, [1.0, 0.0]),
        ({"weights": [0.0, 0.0, 1.0]}, [1.0, 1.0]),
        ({"weights": [0.5, 0.5, 0.0]}, [0.5, 0.0]),
        ({"weights": [0.0, 0.5, 0.5]}, [1.0, 0.5]),
        ({"weights": [0.5, 0.0, 0.5]}, [0.5, 1.0]),
        ({"weights": [1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0]}, [2.0 / 3.0, 0.5]),
        ({"weights": [0.3, 0.3, 0.4]}, [0.7, 4.0 / 7.0]),
        (
            {
                "weights": [
                    np.array([1.0, 0.6, 0.2, 0.2, 0.0]),
                    np.array([0.0, 0.2, 0.6, 0.2, 0.5]),
                    np.array([0.0, 0.2, 0.2, 0.6, 0.5]),
                ],
                "geometry": "plane_parallel",
            },
            [
                [0.0, 0.4, 0.8, 0.8, 1.0],
                [0.0, 0.5, 0.25, 0.75, 0.5],
            ],
        ),
    ],
    ids=[
        "scalar0",
        "scalar1",
        "scalar2",
        "scalar3",
        "scalar4",
        "scalar5",
        "scalar6",
        "scalar7",
        "scalar8",
        "array",
    ],
)
def test_blend_phase_kernel_dict_3_components(mode_mono, kwargs, expected_mi_weights):
    """
    Blendphase with 3 components produces correct kernel dict and can be loaded.
    """
    phase = BlendPhaseFunction(
        components=[
            {"type": "hg", "g": -0.1},
            {"type": "rayleigh"},
            {"type": "hg", "g": 0.1},
        ],
        **kwargs,
    )
    check_scene_element(phase, mi.PhaseFunction)

    template, params = traverse(phase)
    kernel_dict = mi_to_numpy(template.render(KernelContext(), nested=False))

    # Array weights
    expected = {
        "type": "blendphase",
        "weight.type": "gridvolume",
        "weight.grid": np.reshape(expected_mi_weights[0], (-1, 1, 1)),
        "phase_0.type": "hg",
        "phase_0.g": -0.1,
        "phase_1.type": "blendphase",
        "phase_1.weight.type": "gridvolume",
        "phase_1.weight.grid": np.reshape(expected_mi_weights[1], (-1, 1, 1)),
        "phase_1.phase_0.type": "rayleigh",
        "phase_1.phase_1.type": "hg",
        "phase_1.phase_1.g": 0.1,
    }

    if "geometry" in kwargs:
        geometry = SceneGeometry.convert(kwargs["geometry"])
        expected["weight.to_world"] = np.array(
            geometry.atmosphere_volume_to_world.matrix
        )
        expected["phase_1.weight.to_world"] = expected["weight.to_world"].copy()

    assert_cmp_dict(expected, kernel_dict)

    # Check that the parameter map is correct
    assert set(params.keys()) == {
        "weight.data",
        "phase_0.g",
        "phase_1.weight.data",
        "phase_1.phase_1.g",
    }



import numpy as np
import pytest

from eradiate import unit_registry as ureg
from eradiate.frame import (
    angles_in_hplane,
    angles_to_direction,
    cos_angle_to_direction,
    direction_to_angles,
    spherical_to_cartesian,
    transform_azimuth,
)


def test_cos_angle_to_direction():
    # Scalar call
    assert np.allclose(cos_angle_to_direction(1.0, 0.0), [0, 0, 1])
    assert np.allclose(cos_angle_to_direction(0.0, 0.0), [1, 0, 0])
    assert np.allclose(cos_angle_to_direction(0.5, 0.0), [np.sqrt(3) / 2, 0, 0.5])
    assert np.allclose(
        cos_angle_to_direction(-1.0, ureg.Quantity(135.0, "deg")), [0, 0, -1]
    )

    # Vectorised call
    expected = np.array([[0, 0, 1], [1, 0, 0], [np.sqrt(3) / 2, 0, 0.5], [0, 0, -1]])
    assert np.allclose(
        cos_angle_to_direction([1.0, 0.0, 0.5, -1.0], [0, 0, 0.0, 0.75 * np.pi]),
        expected,
    )

    # Reverse parameter returns the opposite vector
    assert np.allclose(
        cos_angle_to_direction(
            [1.0, 0.0, 0.5, -1.0], [0, 0, 0.0, 0.75 * np.pi], flip=True
        ),
        -expected,
    )


@pytest.mark.parametrize(
    ["from_convention", "normalize", "expected"],
    [
        ("east_right", True, [0.0, 45.0, 135.0]),
        ("east_left", True, [0.0, 315.0, 225.0]),
        ("north_right", True, [90.0, 135.0, 225.0]),
        ("north_left", True, [90.0, 45.0, 315.0]),
        ("west_right", True, [180.0, 225.0, 315.0]),
        ("west_left", True, [180.0, 135.0, 45.0]),
        ("south_right", True, [270.0, 315.0, 45.0]),
        ("south_left", True, [270.0, 225.0, 135.0]),
        ("south_right", False, [270.0, 315.0, 405.0]),
    ],
)
def test_transform_azimuth_to_east_right(from_convention, expected, normalize):
    initial = np.deg2rad([0.0, 45.0, 135.0])
    result = transform_azimuth(
        initial,
        from_convention=from_convention,
        normalize=normalize,
    )
    assert np.allclose(np.rad2deg(result), expected)


@pytest.mark.parametrize(
    ["to_convention", "normalize", "initial"],
    [
        ("east_right", True, [0.0, 45.0, 135.0]),
        ("east_left", True, [0.0, 315.0, 225.0]),
        ("north_right", True, [90.0, 135.0, 225.0]),
        ("north_left", True, [90.0, 45.0, 315.0]),
        ("west_right", True, [180.0, 225.0, 315.0]),
        ("west_left", True, [180.0, 135.0, 45.0]),
        ("south_right", True, [270.0, 315.0, 45.0]),
        ("south_left", True, [270.0, 225.0, 135.0]),
        ("south_right", False, [270.0, 315.0, 405.0]),
        # The following configurations test for close-to-zero value snapping
        ("east_right", True, [360.0 - 1e-4, 45.0, 135.0]),
        ("north_right", True, [90.0 - 1e-4, 135.0, 225.0]),
    ],
)
def test_transform_azimuth_from_east_right(to_convention, initial, normalize):
    expected = np.deg2rad([0.0, 45.0, 135.0])
    result = transform_azimuth(
        np.deg2rad(initial),
        to_convention=to_convention,
        normalize=normalize,
    )
    assert np.allclose(result, expected)


def test_angles_to_direction():
    # Scalar call
    assert np.allclose(angles_to_direction([0.0, 0.0]), [0, 0, 1])
    assert np.allclose(angles_to_direction([np.pi, 0.0]), [0, 0, -1])
    assert np.allclose(angles_to_direction([0.5 * np.pi, 0.0]), [1, 0, 0])
    assert np.allclose(angles_to_direction([0.5 * np.pi, np.pi]), [-1, 0, 0])
    assert np.allclose(angles_to_direction([0.5 * np.pi, 0.5 * np.pi]), [0, 1, 0])
    assert np.allclose(angles_to_direction([0.5 * np.pi, -0.5 * np.pi]), [0, -1, 0])
    assert np.allclose(
        angles_to_direction([0.25 * np.pi, 0]), [0.70710678, 0, 0.70710678]
    )
    assert np.allclose(
        angles_to_direction([0.5 * np.pi, 0.25 * np.pi]), [0.70710678, 0.70710678, 0]
    )

    # Vectorised call
    assert np.allclose(
        angles_to_direction(
            [
                [0.0, 0.0],
                [np.pi, 0.0],
                [0.5 * np.pi, 0.0],
                [0.5 * np.pi, np.pi],
                [0.5 * np.pi, 0.5 * np.pi],
                [0.5 * np.pi, -0.5 * np.pi],
                [0.25 * np.pi, 0],
                [0.5 * np.pi, 0.25 * np.pi],
                [-0.25 * np.pi, 0],
            ]
        ),
        [
            [0, 0, 1],
            [0, 0, -1],
            [1, 0, 0],
            [-1, 0, 0],
            [0, 1, 0],
            [0, -1, 0],
            [np.sqrt(2) * 0.5, 0, np.sqrt(2) * 0.5],
            [np.sqrt(2) * 0.5, np.sqrt(2) * 0.5, 0],
            [-np.sqrt(2) * 0.5, 0, np.sqrt(2) * 0.5],
        ],
    )


def test_direction_to_angles():
    # Scalar call
    assert np.allclose(direction_to_angles([0, 0, 1]), (0, 0))
    assert np.allclose(
        direction_to_angles([np.sqrt(2) / 2, np.sqrt(2) / 2, 0]).m_as(ureg.deg),
        [90.0, 45.0],
    )
    assert np.allclose(
        direction_to_angles([1, 1, 0]).m_as(ureg.deg),
        [90.0, 45.0],
    )
    assert np.allclose(
        direction_to_angles(
            [1 / np.sqrt(3), 1 / np.sqrt(3), 1 / np.sqrt(3)],
        ).m_as(ureg.rad),
        [np.arccos(1 / np.sqrt(3)), 0.25 * np.pi],
    )
    assert np.allclose(
        direction_to_angles([1, 1, 1]).m_as(ureg.rad),
        [np.arccos(1 / np.sqrt(3)), 0.25 * np.pi],
    )

    # Vectorised call
    assert np.allclose(
        direction_to_angles(
            [
                [0, 0, 1],
                [1, 0, 1],
                [0, 1, 1],
                [1, 0, 0],
                [0, 1, 0],
                [1, 1, 1],
                [0, 0, -1],
            ]
        ),
        [
            [0, 0],
            [0.25 * np.pi, 0],
            [0.25 * np.pi, 0.5 * np.pi],
            [0.5 * np.pi, 0],
            [0.5 * np.pi, 0.5 * np.pi],
            [np.arccos(1 / np.sqrt(3)), 0.25 * np.pi],
            [np.pi, 0.0],
        ],
    )


def test_spherical_to_cartesian():
    r = 2.0
    theta = np.deg2rad(30)
    phi = np.deg2rad(0)
    d = spherical_to_cartesian(r, theta, phi)
    assert np.allclose(d, [1, 0, np.sqrt(3)])

    r = ureg.Quantity(2.0, "km")
    theta = np.deg2rad(60)
    phi = np.deg2rad(30)
    d = spherical_to_cartesian(r, theta, phi)
    assert np.allclose(d, ureg.Quantity([3.0 / 2.0, np.sqrt(3) / 2.0, 1.0], "km"))


@pytest.mark.parametrize(
    "theta, phi, plane, expected_plane",
    [
        (0, 0, 0, "p"),
        (45, 0, 0, "p"),
        (-45, 0, 0, "n"),
        (90, 0, 0, "p"),
        (-90, 0, 0, "n"),
        (0, 180, 0, "p"),
        (45, 180, 0, "n"),
        (-45, 180, 0, "p"),
        (90, 180, 0, "n"),
        (-90, 180, 0, "p"),
        (0, 0, 180, "p"),
        (45, 0, 180, "n"),
        (-45, 0, 180, "p"),
        (90, 0, 180, "n"),
        (-90, 0, 180, "p"),
        (0, 180, 180, "p"),
        (45, 180, 180, "p"),
        (-45, 180, 180, "n"),
        (90, 180, 180, "p"),
        (-90, 180, 180, "n"),
        (0, 0, 45, "p"),
        (45, 0, 45, None),
        (-45, 0, 45, None),
        (45, 0, -45, None),
        (-45, 0, -45, None),
    ],
)
def test_angles_in_hplane(theta, phi, plane, expected_plane):
    p, n = angles_in_hplane(
        np.deg2rad(plane), np.deg2rad(theta), np.deg2rad(phi), raise_exc=False
    )

    if expected_plane == "p":
        assert p and not n

    if expected_plane == "n":
        assert n and not p

    if expected_plane is None:
        assert not p and not n



"""Root module testing."""
import pytest

import eradiate
from eradiate import supported_mode, unsupported_mode
from eradiate.exceptions import UnsupportedModeError


def test_mode_mono(mode_mono):
    mode = eradiate.mode()
    assert mode.mi_variant == "scalar_mono_double"


def test_mode_mono_single(mode_mono_single):
    mode = eradiate.mode()
    assert mode.mi_variant == "scalar_mono"


def test_mode_mono_double(mode_mono_double):
    mode = eradiate.mode()
    assert mode.mi_variant == "scalar_mono_double"


def test_modes():
    import mitsuba

    # Check that switching to double precision mode works
    eradiate.set_mode("mono_double")
    # We expect that the kernel variant is appropriately selected
    assert mitsuba.variant() == "scalar_mono_double"

    # Check that switching to mono mode works
    eradiate.set_mode("mono_single")
    # We expect that the kernel variant is appropriately selected
    assert mitsuba.variant() == "scalar_mono"


def test_mode_check():
    # Check flags for mono single mode
    eradiate.set_mode("mono_single")
    assert eradiate.mode().check(spectral_mode="mono", mi_double_precision=False)

    # Check flags for mono_double mode
    eradiate.set_mode("mono_double")
    assert eradiate.mode().check(spectral_mode="mono", mi_double_precision=True)


def test_supported_mode():
    for mode in ["mono_single", "mono_double", "mono"]:
        eradiate.set_mode(mode)

        with pytest.raises(UnsupportedModeError):
            supported_mode(spectral_mode="ckd")

        supported_mode(spectral_mode="mono")

    for mode in ["ckd_single", "ckd_double", "ckd"]:
        eradiate.set_mode(mode)

        with pytest.raises(UnsupportedModeError):
            supported_mode(spectral_mode="mono")

        supported_mode(spectral_mode="ckd")

    for mode in ["mono", "ckd"]:
        eradiate.set_mode(mode)

        with pytest.raises(UnsupportedModeError):
            supported_mode(mi_double_precision=False)

        supported_mode(mi_double_precision=True)

    for mode in ["mono_double", "ckd_double"]:
        eradiate.set_mode(mode)

        with pytest.raises(UnsupportedModeError):
            supported_mode(mi_double_precision=False)

        supported_mode(mi_double_precision=True)


def test_unsupported_mode():
    for mode in ["mono_single", "mono_double", "mono"]:
        eradiate.set_mode(mode)

        with pytest.raises(UnsupportedModeError):
            unsupported_mode(spectral_mode="mono")

        unsupported_mode(spectral_mode="ckd")

    for mode in ["ckd_single", "ckd_double", "ckd"]:
        eradiate.set_mode(mode)

        with pytest.raises(UnsupportedModeError):
            unsupported_mode(spectral_mode="ckd")

        unsupported_mode(spectral_mode="mono")

    for mode in ["mono_double", "ckd_double", "mono", "ckd"]:
        eradiate.set_mode(mode)

        with pytest.raises(UnsupportedModeError):
            unsupported_mode(mi_double_precision=True)

        unsupported_mode(mi_double_precision=False)

    for mode in ["mono_single", "ckd_single"]:
        eradiate.set_mode(mode)

        with pytest.raises(UnsupportedModeError):
            unsupported_mode(mi_double_precision=False)

        unsupported_mode(mi_double_precision=True)



import mitsuba as mi
import numpy as np
import pytest

import eradiate
from eradiate import unit_registry as ureg
from eradiate.exceptions import DataError
from eradiate.scenes.spectra import SolarIrradianceSpectrum
from eradiate.spectral.index import SpectralIndex
from eradiate.test_tools.types import check_scene_element
from eradiate.units import PhysicalQuantity


@pytest.mark.parametrize(
    "tested, expected",
    [
        ({}, SolarIrradianceSpectrum()),
        ({"dataset": "doesnt_exist"}, DataError),
    ],
    ids=[
        "no_args",
        "dataset_doesnt_exist",
    ],
)
def test_solar_irradiance_construct(modes_all, tested, expected):
    if isinstance(expected, SolarIrradianceSpectrum):
        s = SolarIrradianceSpectrum(**tested)
        assert s.quantity is PhysicalQuantity.IRRADIANCE

    elif issubclass(expected, Exception):
        with pytest.raises(expected):
            SolarIrradianceSpectrum(**tested)

    else:
        raise RuntimeError


@pytest.mark.parametrize(
    "tested",
    [
        {},
        {"scale": 2.0},
        {"dataset": "solid_2017-mean"},
    ],
    ids=[
        "no_args",
        "scale",
        "solid_spectrum",
    ],
)
def test_solar_irradiance_kernel_dict(mode_mono, tested):
    s = SolarIrradianceSpectrum(**tested)
    check_scene_element(s, mi.Texture)


def test_solar_irradiance_eval(modes_all_double):
    # Irradiance is correctly interpolated in mono mode
    s = SolarIrradianceSpectrum(dataset="thuillier_2003")

    si = SpectralIndex.new(w=550.0 * ureg.nm)
    expected = 1.87938 * ureg.W / ureg.m**2 / ureg.nm  # computed manually

    assert np.allclose(s.eval(si), expected)

    # Eval raises out of the supported spectral range
    if eradiate.mode().is_mono:
        with pytest.raises(ValueError):
            s.eval(SpectralIndex.new(w=1.0 * ureg.km))

    elif eradiate.mode().is_ckd:
        pass

    else:
        assert False


def test_solar_irradiance_scale(mode_mono):
    s = SolarIrradianceSpectrum(dataset="thuillier_2003")

    # We can scale the spectrum using a float
    s_scaled_float = SolarIrradianceSpectrum(dataset="thuillier_2003", scale=10.0)
    assert (
        s_scaled_float.eval_mono(550.0 * ureg.nm) == s.eval_mono(550.0 * ureg.nm) * 10.0
    )


def test_solar_irradiance_datetime(mode_mono):
    s = SolarIrradianceSpectrum(dataset="thuillier_2003")

    # We can also use a datetime to scale the spectrum
    s_scaled_datetime = SolarIrradianceSpectrum(
        dataset="thuillier_2003", datetime="2021-11-18"
    )
    assert np.isclose(
        s_scaled_datetime.eval_mono(550.0 * ureg.nm),
        s.eval_mono(550.0 * ureg.nm) / 0.98854537**2,
    )



import datetime
import os
import sys
import unittest.mock as mock
from importlib.metadata import version

# -- Mock Mitsuba modules (required to render CLI reference on RTD) ------------

MOCK_MODULES = []
for mod_name in MOCK_MODULES:
    sys.modules[mod_name] = mock.Mock()

# -- Path setup ----------------------------------------------------------------

# sys.path.insert(0, os.path.abspath('.'))
sys.path.append(os.path.abspath("./_ext"))

# -- Project information -------------------------------------------------------

project = "Eradiate"
copyright = f"2020-{datetime.datetime.now().year}, The Eradiate Team"
author = "The Eradiate Team"
release = version("eradiate")
version = ".".join(release.split(".")[:3])
if "dev" in release:
    version += ".dev"

# -- General configuration -----------------------------------------------------


html_theme = "sphinx_book_theme"
html_title = ""
html_logo = "_static/eradiate-logo-typo-darkgrey.svg"

templates_path = ["_templates"]  # Path to templates, relative to this directory.
exclude_patterns = ["_build", "tutorials/README.md"]
html_static_path = ["_static"]

extensions = [
    # Core extensions
    "sphinx.ext.autodoc",  # Automatic API documentation
    "sphinx.ext.autosummary",  # Summary tables in API docs
    "sphinx.ext.doctest",  # Doctest blocks
    "sphinx.ext.extlinks",  # External links with dedicated roles
    "sphinx.ext.intersphinx",  # Inter-project links
    "sphinx.ext.mathjax",  # Equation support
    "sphinx.ext.napoleon",  # Better docstrings
    "sphinx.ext.viewcode",  # Add links to highlighted source code
    # Third-party
    "autodocsumm",  # Possibly add autosummary table to autodoc
    "myst_parser",  # Markdown support
    "nbsphinx",  # Display notebooks
    "sphinx_copybutton",
    "sphinx_design",  # Tabs, buttons, icons...
    "sphinxcontrib.bibtex",  # BibTeX bibliography
    # Custom extensions
    "pluginparameters",  # Directives and roles to document Mitsuba plugins
]

# -- GitHub quicklinks with 'extlinks' -----------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/extlinks.html

ghbase = "https://github.com"
ghroot = f"{ghbase}/eradiate/eradiate"
extlinks = {
    "ghissue": (f"{ghroot}/issues/%s", "GH%s"),
    "ghpr": (f"{ghroot}/pull/%s", "PR%s"),
    "ghcommit": (f"{ghroot}/commit/%s", "%.7s"),
    "ghuser": (f"{ghbase}/%s", "@%s"),
}

# -- Bibliography with 'sphinxcontrib.bibtex' ----------------------------------
# https://sphinxcontrib-bibtex.readthedocs.io/en/latest/

bibtex_bibfiles = ["references.bib"]
bibtex_reference_style = "author_year"

# -- Tutorials -----------------------------------------------------------------
# https://nbsphinx.readthedocs.io/

if not os.path.exists("tutorials"):
    os.symlink("../tutorials", "tutorials", target_is_directory=True)
nbsphinx_execute = "never"
nbsphinx_prolog = """
.. button-link:: https://github.com/eradiate/eradiate-tutorials/blob/main/{{ env.doc2path(env.docname, base=False)|replace("tutorials/", "") }}
   :color: primary
   :expand:

   :fas:`external-link-alt` Go to notebook file

----
"""

# -- Intersphinx configuration for cross-project referencing -------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/intersphinx.html#configuration

intersphinx_mapping = {
    "astropy": ("https://docs.astropy.org/en/stable/", None),
    "attr": ("https://www.attrs.org/en/stable/", None),
    "attrs": ("https://www.attrs.org/en/stable/", None),
    "dateutil": ("https://dateutil.readthedocs.io/en/stable/", None),
    "dessinemoi": ("https://dessinemoi.readthedocs.io/en/stable/", None),
    "hamilton": ("https://hamilton.dagworks.io/en/latest/", None),
    "matplotlib": ("https://matplotlib.org/stable/", None),
    "mitsuba": ("https://mitsuba.readthedocs.io/en/stable/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "pandas": ("https://pandas.pydata.org/docs/", None),
    "pooch": ("https://www.fatiando.org/pooch/latest/", None),
    "pint": ("https://pint.readthedocs.io/en/stable/", None),
    "pinttrs": ("https://pinttrs.readthedocs.io/en/stable/", None),
    "python": ("https://docs.python.org/3/", None),
    "rich": ("https://rich.readthedocs.io/en/stable/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/", None),
    "xarray": ("https://docs.xarray.dev/en/stable/", None),
}

# -- Autodoc options -----------------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html

autodoc_default_flags = [
    "members",
    "undoc-members",
    "show-inheritance",
    "inherited-members",
]
autodoc_typehints = "none"
autodoc_member_order = "groupwise"
autodoc_preserve_defaults = False

# Autosummary tables in autodoc
# https://autodocsumm.readthedocs.io/en/latest
autodoc_default_options = {
    # "autosummary": False,
}

# -- Autosummary options -------------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/autosummary.html

autosummary_generate = True
autosummary_members = True

# -- Docstrings ----------------------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/napoleon.html

napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_use_param = False
napoleon_use_rtype = False
napoleon_custom_sections = [("Fields", "params_style")]
napoleon_preprocess_types = True
napoleon_type_aliases = {
    # general terms
    "callable": ":py:func:`callable`",
    "datetime": ":py:class:`~datetime.datetime`",
    "dict_like": ":term:`dict-like <mapping>`",
    "dict-like": ":term:`dict-like <mapping>`",
    "file-like": ":term:`file-like <file-like object>`",
    "iterable": ":term:`iterable`",
    "path-like": ":term:`path-like <path-like object>`",
    "mapping": ":term:`mapping`",
    "sequence": ":term:`sequence`",
    # stdlib type aliases
    "MutableMapping": "~collections.abc.MutableMapping",
    "Path": "~pathlib.Path",
    # numpy terms and aliases
    "array": ":term:`array`",
    "array_like": ":term:`array_like`",
    "array-like": ":term:`array-like <array_like>`",
    "dtype": "~numpy.dtype",
    "hashable": ":term:`hashable <name>`",
    "ndarray": "~numpy.ndarray",
    "scalar": ":term:`scalar`",
    "MaskedArray": "~numpy.ma.MaskedArray",
    # matplotlib terms and aliases
    "matplotlib axes": ":py:class:`matplotlib axes <matplotlib.axes.Axes>`",
    "axes": ":py:class:`axes <matplotlib.axes.Axes>`",
    "matplotlib figure": ":py:class:`matplotlib figure <matplotlib.figure.Figure>`",
    "figure": ":py:class:`figure <matplotlib.figure.Figure>`",
    # xarray terms and aliases
    "DataArray": "~xarray.DataArray",
    "Dataset": "~xarray.Dataset",
    "Variable": "~xarray.Variable",
    # pint terms and aliases
    "quantity": ":class:`quantity <pint.Quantity>`",
    # pandas terms and aliases
    "pd.Index": "~pandas.Index",
    "pd.NaT": "~pandas.NaT",
    # local aliases
    "AUTO": ":data:`~eradiate.attrs.AUTO`",
}

# -- Extra roles (requires the 'pluginparameters' extension) -------------------

rst_prolog = r"""
.. role:: bolditalic
   :class: font-weight-bold font-italic

.. role:: monospace
   :class: text-monospace

.. role:: monobold
   :class: text-monospace font-weight-bold

.. role:: monosp
   :class: text-monospace

.. role:: paramtype
   :class: text-monospace

.. |spectrum| replace:: :paramtype:`spectrum`
.. |texture| replace:: :paramtype:`texture`
.. |float| replace:: :paramtype:`float`
.. |bool| replace:: :paramtype:`boolean`
.. |int| replace:: :paramtype:`integer`
.. |false| replace:: :monosp:`false`
.. |true| replace:: :monosp:`true`
.. |string| replace:: :paramtype:`string`
.. |bsdf| replace:: :paramtype:`bsdf`
.. |phase| replace:: :paramtype:`phase`
.. |point| replace:: :paramtype:`point`
.. |vector| replace:: :paramtype:`vector`
.. |transform| replace:: :paramtype:`transform`
.. |volume| replace:: :paramtype:`volume`
.. |tensor| replace:: :paramtype:`tensor`

.. |exposed| replace:: :abbr:`P (This parameters will be exposed as a scene parameter)`
.. |differentiable| replace:: :abbr:`∂ (This parameter is differentiable)`
.. |discontinuous| replace:: :abbr:`D (This parameter might introduce discontinuities. Therefore it requires special handling during differentiation to prevent bias (e.g. prb-reparam)))`
"""

# -- HTML output customisation -------------------------------------------------

html_theme_options = {
    "repository_url": "https://github.com/eradiate/eradiate",
    "repository_branch": "main",
    "use_repository_button": True,
    "use_issues_button": True,
    "use_sidenotes": True,
    "logo": {
        "image_light": "_static/eradiate-logo-typo-black.svg",
        "image_dark": "_static/eradiate-logo-typo-white.svg",
    },
}

html_css_files = ["theme_overrides.css"]
html_short_title = "Eradiate"
html_favicon = "fig/icon_eradiate.png"
html_show_sourcelink = False
htmlhelp_basename = "eradiate_doc"

# -- LaTeX output customisation ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    "papersize": "a4paper",
    "classoptions": ",oneside",
    # The font size ("10pt", "11pt" or "12pt").
    "pointsize": "10pt",
    # Fonts
    "fontpkg": r"""
        \setmainfont{Charis SIL}[Scale=.98]
        \setsansfont{Source Sans Pro}[Scale=MatchLowercase]
        \setmonofont{Hack}[Scale=MatchLowercase]
    """,
    # Additional stuff for the LaTeX preamble.
    "preamble": "",
    # Latex figure (float) alignment
    # 'figure_align': 'htbp',
}

latex_engine = "xelatex"
latex_logo = "fig/eradiate-logo-typo-black.png"

latex_documents = [
    (
        "index_latex",
        "eradiate.tex",
        "Eradiate Documentation",
        "The Eradiate Team",
        "manual",
    ),
]


# -------------------------- Custom generation steps ---------------------------


def custom_step(app):
    sys.path.append(os.path.dirname(os.path.abspath(__file__)))
    import generate_md_cli
    import generate_rst_api
    import generate_rst_plugins

    generate_md_cli.generate()  # CLI reference
    generate_rst_plugins.generate()  # Plugins
    generate_rst_api.generate_env_vars_docs()  # Environment variables
    generate_rst_api.generate_factory_docs()  # Factories


def setup(app):
    app.connect("builder-inited", custom_step)



from eradiate import data

TEST_FILE_SUBMODULE = "tests/data/git/registered_dataset.nc"
TEST_FILE_ONLINE = "tests/data/online/registered_dataset.nc"


def test_open_dataset():
    # Try opening a dataset from the submodule
    with data.open_dataset(TEST_FILE_SUBMODULE):
        assert True

    # Try opening a very small dataset from the online store
    with data.open_dataset(TEST_FILE_ONLINE):
        assert True


def test_load_dataset():
    assert data.load_dataset(TEST_FILE_SUBMODULE)



from . import absorption as absorption
from . import rayleigh as rayleigh
from ._atmosphere import AtmosphereRadProfile as AtmosphereRadProfile
from ._core import RadProfile as RadProfile
from ._core import ZGrid as ZGrid
from ._core import rad_profile_factory as rad_profile_factory



from itertools import product

import mitsuba as mi
import numpy as np

from eradiate import KernelContext
from eradiate.kernel import (
    MitsubaObjectWrapper,
    TypeIdLookupStrategy,
    UpdateMapTemplate,
    UpdateParameter,
    mi_render,
    mi_traverse,
)
from eradiate.spectral.index import SpectralIndex
from eradiate.units import unit_registry as ureg


def test_type_id_lookup_strategy(mode_mono):
    mi_scene = mi.load_dict(
        {
            "type": "scene",
            "bsdf": {"type": "diffuse", "id": "my_bsdf"},
            "rectangle_1": {
                "type": "rectangle",
                "bsdf": {"type": "ref", "id": "my_bsdf"},
            },
            "rectangle_2": {
                "type": "rectangle",
                "bsdf": {"type": "ref", "id": "my_bsdf"},
            },
            "disk_1": {
                "type": "disk",
                "bsdf": {"type": "ref", "id": "my_bsdf"},
            },
            "disk_2": {
                "type": "disk",
                "bsdf": {"type": "diffuse"},
            },
        }
    )

    lookup_strategy = TypeIdLookupStrategy(
        node_type=mi.BSDF, node_id="my_bsdf", parameter_relpath="reflectance.value"
    )

    for shape in mi_scene.shapes():
        path = shape.id()
        assert lookup_strategy(shape.bsdf(), path) == (
            f"{path}.reflectance.value" if path != "disk_2" else None
        )


def test_mi_traverse(mode_mono):
    mi_scene = mi.load_dict(
        {
            "type": "scene",
            "bsdf": {"type": "diffuse", "id": "my_bsdf"},
            "rectangle_1": {
                "type": "rectangle",
                "bsdf": {"type": "ref", "id": "my_bsdf"},
            },
            "rectangle_2": {
                "type": "rectangle",
                "bsdf": {"type": "ref", "id": "my_bsdf"},
            },
            "disk_1": {
                "type": "disk",
                "bsdf": {"type": "ref", "id": "my_bsdf"},
            },
            "disk_2": {
                "type": "disk",
                "bsdf": {"type": "diffuse"},
            },
        }
    )

    umap_template = UpdateMapTemplate(
        {
            "my_bsdf.reflectance.value": UpdateParameter(
                evaluator=lambda x: x,
                flags=UpdateParameter.Flags.ALL,
                lookup_strategy=TypeIdLookupStrategy(
                    node_type=mi.BSDF,
                    node_id="my_bsdf",
                    parameter_relpath="reflectance.value",
                ),
            )
        }
    )

    mi_wrapper = mi_traverse(mi_scene, umap_template)

    # Traversal succeeds
    assert isinstance(mi_wrapper, MitsubaObjectWrapper)

    # Parameter map is correctly extracted
    assert set(mi_wrapper.parameters.keys()) == {
        "my_bsdf.reflectance.value",
        "disk_1.to_world",
        "disk_2.bsdf.reflectance.value",
        "disk_2.to_world",
        "rectangle_1.to_world",
        "rectangle_2.to_world",
    }

    # Parameter ID should be resolved in accordance with the declared lookup strategy
    assert (
        mi_wrapper.umap_template["my_bsdf.reflectance.value"].parameter_id
        == "my_bsdf.reflectance.value"
    )


def test_mi_render(mode_mono):
    mi_scene = mi.load_dict(
        {
            "type": "scene",
            "rectangle": {
                "type": "rectangle",
                "bsdf": {"type": "diffuse", "id": "my_bsdf"},
            },
            "sensor": {
                "type": "distant",
                "film": {"type": "hdrfilm", "width": 1, "height": 1},
                "direction": [0, 0, -1],
                "target": [0, 0, 0],
            },
            "illumination": {
                "type": "directional",
                "direction": [0, 0, -1],
                "irradiance": 1.0,
            },
            "integrator": {"type": "path"},
        }
    )

    umap_template = UpdateMapTemplate(
        {
            "my_bsdf.reflectance.value": UpdateParameter(
                evaluator=lambda ctx: ctx.kwargs["r"],
                flags=UpdateParameter.Flags.ALL,
                lookup_strategy=TypeIdLookupStrategy(
                    node_type=mi.BSDF,
                    node_id="my_bsdf",
                    parameter_relpath="reflectance.value",
                ),
            )
        }
    )

    mi_wrapper = mi_traverse(mi_scene, umap_template)

    reflectances = [0.0, 0.5, 1.0]
    wavelengths = [400.0, 500.0, 600.0] * ureg.nm

    result = mi_render(
        mi_wrapper,
        ctxs=[
            KernelContext(si=SpectralIndex.new(w=w), kwargs={"r": r})
            for (r, w) in zip(reflectances, wavelengths)
        ],
    )

    assert isinstance(result, dict)

    expected = []
    actual = []
    for _, (r, w) in enumerate(zip(reflectances, wavelengths)):
        siah = SpectralIndex.new(w=w).as_hashable
        assert isinstance(result[siah]["sensor"], mi.Bitmap)
        expected.append(r / np.pi)
        actual.append(np.squeeze(result[siah]["sensor"]))

    np.testing.assert_allclose(actual, expected)


def test_mi_render_multisensor(mode_mono):
    mi_scene = mi.load_dict(
        {
            "type": "scene",
            "rectangle": {
                "type": "rectangle",
                "bsdf": {"type": "diffuse", "id": "my_bsdf"},
            },
            "sensor1": {
                "type": "distant",
                "film": {"type": "hdrfilm", "width": 1, "height": 1},
                "direction": [0, 0, -1],
                "target": [0, 0, 0],
            },
            "sensor2": {
                "type": "distant",
                "film": {"type": "hdrfilm", "width": 1, "height": 1},
                "direction": [0, 0, -1],
                "target": [0, 0, 0],
            },
            "illumination": {
                "type": "directional",
                "direction": [0, 0, -1],
                "irradiance": 1.0,
            },
            "integrator": {"type": "path"},
        }
    )

    umap_template = UpdateMapTemplate(
        {
            "my_bsdf.reflectance.value": UpdateParameter(
                evaluator=lambda ctx: ctx.kwargs["r"],
                flags=UpdateParameter.Flags.ALL,
                lookup_strategy=TypeIdLookupStrategy(
                    node_type=mi.BSDF,
                    node_id="my_bsdf",
                    parameter_relpath="reflectance.value",
                ),
            )
        }
    )

    mi_wrapper = mi_traverse(mi_scene, umap_template)

    reflectances = [0.0, 0.5, 1.0]
    wavelengths = [400.0, 500.0, 600.0] * ureg.nm

    result = mi_render(
        mi_wrapper,
        ctxs=[
            KernelContext(si=SpectralIndex.new(w=w), kwargs={"r": r})
            for (r, w) in zip(reflectances, wavelengths)
        ],
    )

    # The result must be a nested dict with one level-one element per wavelength,
    # and one level-two element per sensor
    assert isinstance(result, dict)
    assert np.allclose(
        list(result.keys()),
        [w.m for w in wavelengths],
    )

    sensors_keys = set()
    for spectral_key in result.keys():
        sensors_keys.update(set(result[spectral_key].keys()))
    assert sensors_keys == {"sensor1", "sensor2"}
    assert all(
        isinstance(result[spectral_key][sensor_key], mi.Bitmap)
        for (spectral_key, sensor_key) in product(result.keys(), sensors_keys)
    )



import os

import numpy as np
import pytest
import xarray as xr

import eradiate
from eradiate import unit_registry as ureg
from eradiate.units import symbol

eradiate.plot.set_style()

# ------------------------------------------------------------------------------
#               Customizable output dir for test artifacts
# ------------------------------------------------------------------------------


def pytest_addoption(parser):
    eradiate_source_dir = os.environ.get("ERADIATE_SOURCE_DIR", ".")
    parser.addoption(
        "--artefact-dir",
        action="store",
        default=os.path.join(eradiate_source_dir, "build/test_artefacts/"),
    )


# See: https://stackoverflow.com/a/55301318/3645374
@pytest.fixture(scope="session")
def artefact_dir(pytestconfig):
    option_value = pytestconfig.getoption("artefact_dir")

    if not os.path.isdir(option_value):
        os.makedirs(option_value)

    return option_value


# ------------------------------------------------------------------------------
#                              Other configuration
# ------------------------------------------------------------------------------


def pytest_configure(config):
    markexpr = config.getoption("markexpr", "False")
    has_slow = "not slow" not in markexpr
    has_regression = "not regression" not in markexpr

    if has_slow:
        print(
            "\033[93m"
            "Running slow tests. To skip them, please run "
            "'pytest -m \"not slow\"' "
            "\033[0m"
        )

    if has_regression:
        print(
            "\033[93m"
            "Running regression tests. To skip them, please run "
            "'pytest -m \"not regression\"' "
            "\033[0m"
        )

    config.addinivalue_line(
        "markers", "slow: marks tests as slow (deselect with -m 'not slow')"
    )

    config.addinivalue_line(
        "markers",
        "regression: marks tests as potentially very slow regression tests "
        "(deselect with -m 'not regression')",
    )


@pytest.fixture(scope="session")
def session_timestamp():
    from datetime import datetime

    return datetime.now()


@pytest.fixture
def absorbing_only():
    """Absorbing only particles radiative properties data set fixture."""
    mu = np.linspace(-1.0, 1.0) * ureg.dimensionless
    w = np.linspace(279.0, 2401.0) * ureg.nm
    arrays = [np.ones_like(mu) / ureg.steradian for _ in w]
    phase = np.stack(arrays, axis=0).reshape(w.size, mu.size, 1, 1)
    albedo = np.zeros_like(w) * ureg.dimensionless
    sigma_t = np.ones_like(w) / ureg.km
    return to_dataset(albedo=albedo, sigma_t=sigma_t, phase=phase, mu=mu, w=w)


@pytest.fixture
def scattering_only():
    """Scattering only particles radiative properties data set fixture."""
    mu = np.linspace(-1.0, 1.0) * ureg.dimensionless
    w = np.linspace(279.0, 2401.0) * ureg.nm
    arrays = [np.ones_like(mu) / ureg.steradian for _ in w]
    phase = np.stack(arrays, axis=0).reshape(w.size, mu.size, 1, 1)
    albedo = np.ones_like(w) * ureg.dimensionless
    sigma_t = np.ones_like(w) / ureg.km
    return to_dataset(albedo=albedo, sigma_t=sigma_t, phase=phase, mu=mu, w=w)


@pytest.fixture
def test_particles_dataset():
    """
    Particles radiative properties data set fixture.

    Particle radiative properties are spectrally constant with values specified
    below:
        * the phase function is isotropic
        * the extinction coefficient is 1 / km
        * the albedo is 0.8
    """
    mu = np.linspace(-1.0, 1.0) * ureg.dimensionless
    w = np.linspace(279.0, 2401.0) * ureg.nm
    arrays = [np.ones_like(mu) / ureg.steradian for _ in w]
    phase = np.stack(arrays, axis=0).reshape(w.size, mu.size, 1, 1)
    albedo = 0.8 * np.ones_like(w) * ureg.dimensionless
    sigma_t = np.ones_like(w) / ureg.km
    return to_dataset(albedo=albedo, sigma_t=sigma_t, phase=phase, mu=mu, w=w)


def to_dataset(albedo, sigma_t, phase, mu, w):
    return xr.Dataset(
        data_vars={
            "sigma_t": (
                "w",
                sigma_t.magnitude,
                dict(
                    standard_name="air_volume_extinction_coefficient",
                    units=symbol(sigma_t.units),
                ),
            ),
            "albedo": (
                "w",
                albedo.magnitude,
                dict(
                    standard_name="single_scattering_albedo",
                    units=symbol(albedo.units),
                ),
            ),
            "phase": (
                ("w", "mu", "i", "j"),
                phase.magnitude,
                dict(
                    standard_name="scattering_phase_matrix",
                    units=symbol(phase.units),
                ),
            ),
        },
        coords={
            "w": ("w", w.magnitude, dict(units=symbol(w.units))),
            "mu": (
                "mu",
                mu.magnitude,
                dict(
                    standard_name="scattering_angle_cosine",
                    units=symbol(mu.units),
                ),
            ),
            "i": ("i", [0]),
            "j": ("j", [0]),
        },
    )



import numpy as np
import pytest

from eradiate.scenes.spectra import InterpolatedSpectrum, MultiDeltaSpectrum
from eradiate.spectral.mono import WavelengthSet
from eradiate.units import unit_registry as ureg


def test_select_with_multi_delta():
    """
    Unit tests for :meth:`.WavelengthSet.select_with_srf`.
    """

    wset = WavelengthSet(wavelengths=np.arange(280.0, 2400.0, 1.0) * ureg.nm)

    srf = MultiDeltaSpectrum(wavelengths=550.0 * ureg.nm)
    assert np.isclose(srf.select_in(wset).wavelengths, srf.wavelengths)

    srf = MultiDeltaSpectrum(wavelengths=555.5 * ureg.nm)
    assert np.isclose(srf.select_in(wset).wavelengths, srf.wavelengths)

    srf = MultiDeltaSpectrum(wavelengths=np.linspace(500.0, 600.0) * ureg.nm)
    assert np.allclose(srf.select_in(wset).wavelengths, srf.wavelengths)


@pytest.mark.parametrize(
    "srf, expected_wavelengths",
    [
        (
            InterpolatedSpectrum(
                wavelengths=np.linspace(500.0, 600.0, 11) * ureg.nm,
                values=np.array([0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0], dtype=float)
                * ureg.dimensionless,
            ),
            np.arange(501.0, 580.0, 1.0) * ureg.nm,
        ),
        (
            InterpolatedSpectrum(
                wavelengths=np.linspace(500.5, 600.5, 11) * ureg.nm,
                values=np.array([0, 1, 1, 0, 1, 1, 1, 1, 0, 0, 0], dtype=float)
                * ureg.dimensionless,
            ),
            np.arange(501.0, 581.0, 1.0) * ureg.nm,
        ),
        (
            InterpolatedSpectrum(
                wavelengths=np.linspace(500.0, 600.0, 11) * ureg.nm,
                values=np.array([0, 1, 1, 0, 0, 1, 1, 1, 0, 0, 0], dtype=float)
                * ureg.dimensionless,
            ),
            np.concatenate(
                [
                    np.arange(501.0, 530.0, 1.0),
                    np.arange(541.0, 580.0, 1.0),
                ]
            )
            * ureg.nm,
        ),
    ],
)
def test_select_in_interpolated_spectrum(srf, expected_wavelengths):
    """
    Unit tests for :meth:`eradiate.scenes.spectra.InterpolatedSpectrum.select_in`.
    """
    wset = WavelengthSet(wavelengths=np.arange(280.0, 2400.0, 1.0) * ureg.nm)
    new_wset = srf.select_in(wset)
    assert np.allclose(new_wset.wavelengths, expected_wavelengths)



import numpy as np
import pytest

from eradiate.quad import Quad


@pytest.mark.parametrize(
    "n, exp_nodes, exp_weights",
    [
        (
            1,
            [0],
            [2],
        ),
        (
            2,
            [-np.sqrt(1.0 / 3.0), np.sqrt(1.0 / 3.0)],
            [1, 1],
        ),
        (
            3,
            [-np.sqrt(3.0 / 5.0), 0, np.sqrt(3.0 / 5.0)],
            [5.0 / 9.0, 8.0 / 9.0, 5.0 / 9.0],
        ),
        (
            4,
            [-0.861136, -0.339981, 0.339981, 0.861136],
            [0.347855, 0.652145, 0.652145, 0.347855],
        ),
    ],
    ids=list([f"n-{i}" for i in range(1, 5)]),
)
def test_gauss_legendre(n, exp_nodes, exp_weights, mode_mono):
    quad = Quad.gauss_legendre(n)
    assert np.allclose(quad.nodes, exp_nodes)
    assert np.allclose(quad.weights, exp_weights)


@pytest.mark.parametrize(
    "n, exp_nodes, exp_weights",
    [
        (
            2,
            [-1, 1],
            [1.0, 1.0],
        ),
        (
            3,
            [-1, 0, 1],
            [1.0 / 3.0, 4.0 / 3.0, 1.0 / 3.0],
        ),
        (
            4,
            [-1, -np.sqrt(1.0 / 5.0), np.sqrt(1.0 / 5.0), 1],
            [1.0 / 6.0, 5.0 / 6.0, 5.0 / 6.0, 1.0 / 6.0],
        ),
        (
            5,
            [-1, -np.sqrt(3.0 / 7.0), 0, np.sqrt(3.0 / 7.0), 1],
            [1.0 / 10.0, 49.0 / 90.0, 32.0 / 45.0, 49.0 / 90.0, 1.0 / 10.0],
        ),
    ],
    ids=list([f"n-{i}" for i in range(2, 6)]),
)
def test_gauss_lobatto(n, exp_nodes, exp_weights, mode_mono):
    quad = Quad.gauss_lobatto(n)
    assert np.allclose(quad.nodes, exp_nodes)
    assert np.allclose(quad.weights, exp_weights)


def test_quad_integrate(modes_all):
    quad = Quad.gauss_legendre(10)

    def f(x):
        return x**2

    values = np.array([f(x) for x in quad.eval_nodes(interval=(0, 1))])
    assert np.allclose(quad.integrate(values, interval=(0, 1)), 1.0 / 3.0)



import mitsuba as mi
import numpy as np
import pytest

from eradiate import unit_registry as ureg
from eradiate.scenes.illumination import AstroObjectIllumination
from eradiate.scenes.spectra import SolarIrradianceSpectrum, UniformSpectrum
from eradiate.test_tools.types import check_scene_element


@pytest.mark.parametrize(
    "kwargs, expected_irradiance_type, angular_diameter",
    [({}, SolarIrradianceSpectrum, 1.0), ({"irradiance": 1.0}, UniformSpectrum, 2.0)],
    ids=["noargs", "from_scalar"],
)
def test_astro_object_construct(
    modes_all, kwargs, expected_irradiance_type, angular_diameter
):
    # Construction without argument succeeds
    illumination = AstroObjectIllumination(**kwargs, angular_diameter=angular_diameter)
    assert illumination
    assert isinstance(illumination.irradiance, expected_irradiance_type)
    assert illumination.angular_diameter == angular_diameter * ureg.deg


@pytest.mark.parametrize(
    "zenith, azimuth, direction, to_world",
    [
        (
            0.0 * ureg.deg,
            0.0 * ureg.deg,
            [0, 0, 1],
            [
                [0, 1, 0, 0],
                [-1, 0, 0, 0],
                [0, 0, 1, 0],
                [0, 0, 0, 1],
            ],
        ),
        (
            30 * ureg.deg,
            0.0 * ureg.deg,
            [0.5, 0.0, np.sqrt(3) / 2],
            [
                [0.0, np.sqrt(3) / 2, 0.5, 0.0],
                [-1.0, 0.0, 0.0, 0.0],
                [0.0, -0.5, np.sqrt(3) / 2, 0.0],
                [0.0, 0.0, 0.0, 1.0],
            ],
        ),
    ],
)
def test_astro_object_to_world(mode_mono, zenith, azimuth, direction, to_world):
    """
    Direction is correctly converted to transformation matrix.
    """
    illumination = AstroObjectIllumination(zenith=zenith, azimuth=azimuth)
    np.testing.assert_allclose(illumination.direction, direction)
    np.testing.assert_allclose(illumination._to_world.matrix, to_world)


def test_astro_object_kernel_dict(modes_all_double):
    # The associated kernel dict is correctly formed and can be loaded
    illumination = AstroObjectIllumination()
    check_scene_element(illumination, mi_cls=mi.Emitter)


COS_PI_4 = 0.5 * np.sqrt(2)


@pytest.mark.parametrize(
    "azimuth_convention, expected",
    [
        ("east_right", [0, COS_PI_4, COS_PI_4]),
        ("east_left", [0, -COS_PI_4, COS_PI_4]),
        ("north_right", [-COS_PI_4, 0, COS_PI_4]),
        ("north_left", [COS_PI_4, 0, COS_PI_4]),
        ("west_right", [0, -COS_PI_4, COS_PI_4]),
        ("west_left", [0, COS_PI_4, COS_PI_4]),
        ("south_right", [COS_PI_4, 0, COS_PI_4]),
        ("south_left", [-COS_PI_4, 0, COS_PI_4]),
    ],
)
def test_astro_object_azimuth_convention(mode_mono, azimuth_convention, expected):
    illumination = AstroObjectIllumination(
        zenith=45 * ureg.deg,
        azimuth=90 * ureg.deg,
        azimuth_convention=azimuth_convention,
    )
    assert np.allclose(illumination.direction, expected), illumination.direction



import os

import matplotlib.pyplot as plt
import numpy as np
import pytest

import eradiate
from eradiate.experiments import AtmosphereExperiment
from eradiate.units import unit_registry as ureg


@pytest.mark.slow
def test_albedo(mode_mono, artefact_dir):
    """
    Albedo
    ======

    This system test verifies the behaviour of experiments capable of albedo
    computation.

    Rationale
    ---------

    We use a scene consisting of a single surface with a diffuse, spectrally
    non-uniform BRDF.

    * Geometry: A Lambertian surface with linearly varying reflectance for
      0 to 1 between 500 and 700 nm.
    * Illumination: Directional illumination from the zenith (default irradiance)
      or constant illumination (default radiance).
    * Atmosphere/canopy: No atmosphere nor canopy.
    * Measure: Distant albedo measure with a film of size 64 x 64. This
      guarantees reasonable stratification of the film sampling and ensures
      quick converge to the expected value, thus allowing for a low sample
      count.

    The test is run for the ``AtmosphereExperiment`` and ``CanopyExperiment``
    classes.

    Expected behaviour
    ------------------

    We expect the albedo to be equal to the reflectance of the surface.

    Results
    -------

    .. image:: generated/plots/test_albedo_atmosphere_directional.png
       :width: 100%

    .. image:: generated/plots/test_albedo_canopy_directional.png
       :width: 100%

    .. image:: generated/plots/test_albedo_atmosphere_constant.png
       :width: 100%

    .. image:: generated/plots/test_albedo_canopy_constant.png
       :width: 100%

    """
    exps = {
        "atmosphere_directional": AtmosphereExperiment(
            measures=[
                {
                    "type": "distant_flux",
                    "srf": {
                        "type": "multi_delta",
                        "wavelengths": [500.0, 550.0, 600.0, 650.0, 700.0] * ureg.nm,
                    },
                    "film_resolution": (64, 64),
                    "spp": 256,
                }
            ],
            atmosphere=None,
            surface={
                "type": "lambertian",
                "reflectance": {
                    "type": "interpolated",
                    "wavelengths": [500.0, 700.0],
                    "values": [0.0, 1.0],
                },
            },
            illumination={"type": "directional", "zenith": 0.0},
        ),
        "atmosphere_constant": AtmosphereExperiment(
            measures=[
                {
                    "type": "distant_flux",
                    "srf": {
                        "type": "multi_delta",
                        "wavelengths": [500.0, 550.0, 600.0, 650.0, 700.0] * ureg.nm,
                    },
                    "film_resolution": (64, 64),
                    "spp": 256,
                }
            ],
            atmosphere=None,
            surface={
                "type": "lambertian",
                "reflectance": {
                    "type": "interpolated",
                    "wavelengths": [500.0, 700.0],
                    "values": [0.0, 1.0],
                },
            },
            illumination={"type": "constant"},
        ),
        "canopy_directional": eradiate.experiments.CanopyExperiment(
            measures=[
                {
                    "type": "distant_flux",
                    "srf": {
                        "type": "multi_delta",
                        "wavelengths": [500.0, 550.0, 600.0, 650.0, 700.0] * ureg.nm,
                    },
                    "film_resolution": (64, 64),
                    "spp": 256,
                }
            ],
            canopy=None,
            surface={
                "type": "lambertian",
                "reflectance": {
                    "type": "interpolated",
                    "wavelengths": [500.0, 700.0],
                    "values": [0.0, 1.0],
                },
            },
            illumination={"type": "directional", "zenith": 0.0},
        ),
        "canopy_constant": eradiate.experiments.CanopyExperiment(
            measures=[
                {
                    "type": "distant_flux",
                    "srf": {
                        "type": "multi_delta",
                        "wavelengths": [500.0, 550.0, 600.0, 650.0, 700.0] * ureg.nm,
                    },
                    "film_resolution": (64, 64),
                    "spp": 256,
                }
            ],
            canopy=None,
            surface={
                "type": "lambertian",
                "reflectance": {
                    "type": "interpolated",
                    "wavelengths": [500.0, 700.0],
                    "values": [0.0, 1.0],
                },
            },
            illumination={"type": "directional", "zenith": 0.0},
        ),
    }

    for exp_name, exp in exps.items():
        # Run simulation
        results = eradiate.run(exp)

        # Plot results
        fig, [ax1, ax2] = plt.subplots(1, 2, figsize=(10, 5))
        wavelengths = results["albedo"].w.values
        albedos = results["albedo"].values.squeeze()
        expected = np.array([0.0, 0.25, 0.5, 0.75, 1.0])

        ax1.plot(wavelengths, albedos, linestyle="--", marker="o")
        ax1.set_title("Albedo")
        ax1.set_xlabel("Wavelength [nm]")

        rdiffs = np.divide(
            (albedos - expected),
            expected,
            where=expected != 0.0,
            out=np.zeros_like(expected),
        )
        ax2.plot(wavelengths, rdiffs, linestyle="--", marker="o")
        ax2.set_title("Relative difference")
        ax2.set_xlabel("Wavelength [nm]")
        rdiffs_max = np.max(np.abs(rdiffs[~np.isnan(rdiffs)]))
        exp = np.ceil(np.log10(rdiffs_max))
        ax2.set_xlim(ax1.get_xlim())
        ax2.set_ylim([-(10**exp), 10**exp])
        ax2.yaxis.set_label_position("right")
        ax2.yaxis.tick_right()
        ax2.ticklabel_format(axis="y", style="sci", scilimits=[-3, 3])
        # Hide offset label and add it as axis label
        if abs(exp) >= 3:
            ax2.yaxis.offsetText.set_visible(False)
            ax2.yaxis.set_label_text(f"×$10^{{{int(exp)}}}$")

        plt.suptitle(f"Case: {exp_name}")
        plt.tight_layout()

        filename = f"test_albedo_{exp_name}.png"
        outdir = os.path.join(artefact_dir, "plots")
        os.makedirs(outdir, exist_ok=True)
        fname_plot = os.path.join(outdir, filename)

        fig.savefig(fname_plot, dpi=200)
        plt.close()

        # Check results
        assert np.allclose(np.squeeze(results["albedo"].values), expected, atol=1e-3)



import xarray as xr

import eradiate.kernel.logging
from eradiate.experiments import AtmosphereExperiment
from eradiate.units import unit_registry as ureg

eradiate.kernel.install_logging()


def test_run_function(modes_all_double):
    measure = {
        "type": "mdistant",
        "srf": {
            "type": "multi_delta",
            "wavelengths": [540, 550] * ureg.nm,
        },
    }

    exp = AtmosphereExperiment(atmosphere=None, measures=measure)
    result = eradiate.run(exp)
    assert isinstance(result, xr.Dataset)



import os

import click
from _utils import resolve_package_manager
from ruamel.yaml import YAML


@click.command()
@click.option(
    "-s",
    "--sections",
    default="dependencies,main,recommended,tests,dev,docs,optional",
    help="Dependency sections to include in the produced environment.yml file. "
    "Default: 'dependencies,main,recommended,tests,dev,docs,optional'",
)
@click.option(
    "-o",
    "--output-dir",
    default="./requirements/pip",
    help="Path to output directory. Default: ./requirements/pip",
)
@click.option(
    "-l",
    "--layered-config",
    default="./requirements/layered.yml",
    help="Path to layered requirement dependency configuration file. "
    "Default: ./requirements/layered.yml",
)
@click.option("-q", "--quiet", is_flag=True, help="Suppress terminal output.")
def cli(sections, output_dir, layered_config, quiet):
    # Load dependency list

    sections = [x.strip() for x in sections.split(",")]

    # Load layered dependency dependencies
    yaml = YAML(typ="safe")
    with open(layered_config) as f:
        layered_yml = resolve_package_manager(yaml.load(f.read()), "pip")

    for section in sections:
        if not quiet:
            print(f"Processing section '{section}'")

        packages = layered_yml[section].get("packages", [])

        if not quiet:
            print(f"Writing to {os.path.join(output_dir, f'{section}.in')}")

        # Create .in file
        with open(os.path.join(output_dir, f"{section}.in"), "w") as f:
            # Prepend layered requirement includes
            includes = layered_yml[section].get("includes", [])
            for include in includes:
                f.write(f"-r {include}.in\n")

            constraints = layered_yml[section].get("constraints", [])
            for constraint in constraints:
                f.write(f"-c {constraint}.lock.txt\n")

            f.write("\n".join(packages))
            f.write("\n")


if __name__ == "__main__":
    cli()



from importlib.metadata import PackageNotFoundError, version
from pathlib import Path

import pytest
from environ.exceptions import ConfigError

from eradiate._config import EradiateConfig, ProgressLevel


def eradiate_mitsuba_status(expect_installed=False):
    """
    Check if eradiate-mitsuba is installed in order to skip tests depending
    on the setup type (production or development).
    """

    try:
        version("eradiate-mitsuba")
        return expect_installed
    except PackageNotFoundError:
        return not expect_installed


@pytest.mark.skipif(
    eradiate_mitsuba_status(True),
    reason="This is a production environment: eradiate-mitsuba is installed",
)
def test_development_config(tmpdir):
    # Create an appropriate file tree for tests
    tmpdir_path = Path(tmpdir)
    (tmpdir_path / "eradiate/src/eradiate").mkdir(parents=True)
    (tmpdir_path / "eradiate/src/eradiate/__init__.py").touch()

    # Raises if ERADIATE_SOURCE_DIR and eradiate_mitsuba are missing
    with pytest.raises(ConfigError):
        EradiateConfig.from_environ({})

    # When ERADIATE_SOURCE_DIR points to a correct path
    # (i.e. contains eradiate.__init__.py), config init succeeds
    assert EradiateConfig.from_environ(
        {"ERADIATE_SOURCE_DIR": tmpdir_path / "eradiate"}
    )

    # Otherwise it raises
    with pytest.raises(ConfigError):
        EradiateConfig.from_environ({"ERADIATE_SOURCE_DIR": tmpdir_path})


@pytest.mark.skipif(
    eradiate_mitsuba_status(),
    reason="This is not a production environment: eradiate-mitsuba is not installed",
)
def test_production_config(tmpdir):
    # Create an appropriate file tree for tests
    tmpdir_path = Path(tmpdir)
    (tmpdir_path / "eradiate/src/eradiate").mkdir(parents=True)
    (tmpdir_path / "eradiate/src/eradiate/__init__.py").touch()

    # Should not fail, should detect eradiate-mitsuba instead
    config = EradiateConfig.from_environ({})
    assert config.source_dir is None

    # When ERADIATE_SOURCE_DIR points to a correct path
    # (i.e. contains eradiate.__init__.py), config init succeeds
    # This mixed setup is supported, but should give precedence to ???
    assert EradiateConfig.from_environ(
        {"ERADIATE_SOURCE_DIR": tmpdir_path / "eradiate"}
    )


def test_config_progress():
    cfg = EradiateConfig.from_environ()

    # We can set progress level using a level
    cfg.progress = ProgressLevel.NONE
    assert cfg.progress is ProgressLevel.NONE

    # We can set progress level using an int
    cfg.progress = 0
    assert cfg.progress is ProgressLevel.NONE

    # We can set progress using a string
    cfg.progress = "NONE"
    assert cfg.progress is ProgressLevel.NONE

    # Progress levels are comparable to int
    assert cfg.progress < 1
    assert cfg.progress == 0



import numpy as np

from eradiate.rng import SeedState


def test_seed_state_construct():
    # A SeedState instance can be constructed without argument ...
    SeedState()

    # ... or from an integer seed
    SeedState(0)

    # ... or from an already initialized SeedSequence instance
    SeedState(np.random.SeedSequence(0))


def test_seed_state_generate():
    # We can generate pseudo random seeds using
    s = SeedState(0)
    n1 = s.next().squeeze()
    n2 = s.next().squeeze()
    assert n1 != n2

    # Seed arrays are also available
    assert s.next(10).shape == (10,)

    # We can reset the seed state to restart the seed sequence
    s.reset()
    assert s.next().squeeze() == n1

    # Resetting the seed state with a different seed changes the seed sequence
    s.reset(1)
    assert s.next().squeeze() != n1


def test_seed_state_rng():
    # We can conveniently request seeded RNGs from a SeedState instance
    s = SeedState()
    assert isinstance(s.numpy_default_rng(), np.random.Generator)



import numpy as np
import pytest
import xarray as xr


@pytest.fixture
def dataarray_without_metadata():
    return xr.DataArray(
        data=np.random.random((1, 1, 1, 1, 1)),
        dims=["vza", "vaa", "sza", "saa", "w"],
        coords={
            "vza": [0.0],
            "vaa": [0.0],
            "sza": [0.0],
            "saa": [0.0],
            "w": [500.0],
        },
    )


@pytest.fixture
def dataset_without_metadata():
    return xr.Dataset(
        data_vars={
            "p": (("z_layer", "z_level", "species"), np.random.random((1, 1, 1))),
        },
        coords={
            "z_layer": ("z_layer", [0.0]),
            "z_level": ("z_level", [0.0]),
            "species": ("species", ["foo"]),
        },
    )



from __future__ import annotations

from pathlib import Path

import pandas as pd
import pint

import eradiate

from ..exceptions import DataError


def locate_absorption_data(
    codename: str, mode: str, wavelength_range: pint.Quantity
) -> list[Path]:
    """
    Locate absorption data for a given wavelength range.

    Returns
    -------
    list of pathlib.Path
        List of paths to absorption datasets.

    Raises
    ------
    ValueError
        If no data is available for the requested wavelength range.

    Notes
    -----
    Determine from the spectral mode, dataset codename and wavelength range,
    which absorption datasets are required to cover the requested wavelength
    range. This is done by reading an index file from the remote data store.
    Datasets that have not already been downloaded will be downloaded from the
    remote data store.
    """
    base_path = f"spectra/absorption/{mode}/{codename}"
    try:
        index_path = eradiate.data.data_store.fetch(f"{base_path}/index.csv")
        df = pd.read_csv(index_path)

        # select all files whose associated wavelength range overlaps with the
        # requested wavelength range
        wl_min_units, wl_max_units = None, None
        wl_min_column = None
        wl_max_column = None
        for c in df.columns:
            if c.startswith("wl_min"):
                wl_min_column = c
                # units are enclosed in []
                wl_min_units = c.split("[")[1].split("]")[0]

            if c.startswith("wl_max"):
                wl_max_column = c
                # units are enclosed in []
                wl_max_units = c.split("[")[1].split("]")[0]

        if wl_min_units is None or wl_max_units is None:
            raise DataError(
                "index file must contain at least one column whose name starts "
                "with 'wl_min' and one column whose name starts with 'wl_max'"
            )

        requested_min = wavelength_range.min().m_as(wl_min_units)
        requested_max = wavelength_range.max().m_as(wl_max_units)

        data_wl_min = df[wl_min_column].min()
        data_wl_max = df[wl_max_column].max()

        if requested_min < data_wl_min or requested_max > data_wl_max:
            raise ValueError(
                f"requested wavelength range {wavelength_range} is outside "
                f"the range of the available data ({data_wl_min} {wl_min_units}"
                f" - {data_wl_max} {wl_max_units})"
            )

        filenames = (
            df["filename"]
            .where(
                (df[wl_max_column] >= requested_min)
                & (df[wl_min_column] <= requested_max)
            )
            .dropna()
            .values.tolist()
        )

        if len(filenames) == 0:
            raise ValueError(f"no files found for wavelength range {wavelength_range}")

        return [
            eradiate.data.data_store.fetch(f"{base_path}/{filename}")
            for filename in filenames
        ]

    except DataError:
        # assume no index file is required because there is only one file
        return [eradiate.data.data_store.fetch(f"{base_path}/{codename}.nc")]



import os
from pathlib import Path

import numpy as np
import pytest

import eradiate
from eradiate import unit_registry as ureg
from eradiate.util.misc import (
    Singleton,
    cache_by_id,
    camel_to_snake,
    deduplicate,
    fullname,
    is_vector3,
    natsorted,
)


def test_singleton():
    class MySingleton(metaclass=Singleton):
        pass

    my_singleton1 = MySingleton()
    my_singleton2 = MySingleton()
    assert my_singleton1 is my_singleton2


@pytest.mark.parametrize(
    "value, expected",
    [
        ("aaa", False),
        ([0, 1], False),
        ([0, 2], False),
        ([0, 1, 2], True),
        ([0.0, 1, 2], True),
        ([0, 1, "2"], False),
        (np.array([0, 1, 2]), True),
        (np.array([0, 1]), False),
        (np.array(["0", 1, 2]), False),
        (ureg.Quantity([0, 1, 2], "m"), True),
    ],
)
def test_is_vector3(value, expected):
    result = is_vector3(value)
    assert result == expected


def test_natsort():
    assert natsorted(["10", "1.2", "9"]) == ["1.2", "9", "10"]
    assert natsorted(["1.2", "a1", "9"]) == ["1.2", "9", "a1"]


def test_deduplicate():
    assert deduplicate([1, 1, 2, 3], preserve_order=True) == [1, 2, 3]
    assert deduplicate([2, 1, 3, 1], preserve_order=True) == [2, 1, 3]
    assert deduplicate([2, 1, 3, 1], preserve_order=False) == [1, 2, 3]
    # Note: this latter test may not be reproducible, we might have to change it


def test_camel_to_snake():
    assert camel_to_snake("SomeKindOfThing") == "some_kind_of_thing"


def test_fullname(mode_mono):
    # Functions
    if os.name == "nt":
        assert fullname(os.path.join) == "ntpath.join"
    else:
        assert fullname(os.path.join) == "posixpath.join"

    assert (
        fullname(eradiate.kernel.gridvolume.write_binary_grid3d)
        == "eradiate.kernel.gridvolume.write_binary_grid3d"
    )
    assert (
        fullname(eradiate.kernel.bitmap_to_dataset)
        == "eradiate.kernel._bitmap.bitmap_to_dataset"
    )

    # Methods
    # -- Instance method from class definition
    assert fullname(Path.is_file) == "pathlib.Path.is_file"
    # -- Instance method from instance
    path = Path()
    assert fullname(path.is_file) == "pathlib.Path.is_file"
    # -- Class method from class definition
    assert (
        fullname(eradiate.spectral.SpectralIndex.new)
        == "eradiate.spectral.index.SpectralIndex.new"
    )
    assert (
        fullname(eradiate.spectral.SpectralIndex.new().new)
        == "eradiate.spectral.index.SpectralIndex.new"
    )

    # Classes
    assert fullname(Path) == "pathlib.Path"
    assert (
        fullname(eradiate.scenes.spectra.Spectrum)
        == "eradiate.scenes.spectra._core.Spectrum"
    )


def test_cache_by_id(capsys):
    # Function
    @cache_by_id
    def f(x, y):
        print("Calling f")
        return x, y

    assert f(1, 1) == (1, 1)
    captured = capsys.readouterr()
    assert captured.out == "Calling f\n"
    assert f(1, 1) == (1, 1)
    captured = capsys.readouterr()
    assert captured.out == ""

    assert f(1, 2) == (1, 2)
    captured = capsys.readouterr()
    assert captured.out == "Calling f\n"
    assert f(1, 2) == (1, 2)
    captured = capsys.readouterr()
    assert captured.out == ""

    # Class
    class MyClass:
        @cache_by_id
        def f(self, x, y):
            print("Calling f")
            return x, y

    obj = MyClass()

    assert obj.f(1, 1) == (1, 1)
    captured = capsys.readouterr()
    assert captured.out == "Calling f\n"
    assert obj.f(1, 1) == (1, 1)
    captured = capsys.readouterr()
    assert captured.out == ""

    assert obj.f(1, 2) == (1, 2)
    captured = capsys.readouterr()
    assert captured.out == "Calling f\n"
    assert obj.f(1, 2) == (1, 2)
    captured = capsys.readouterr()
    assert captured.out == ""



import os
from textwrap import dedent

import check_conda_env


def main():
    check_conda_env.main()

    root_dir = os.getcwd()
    conda_prefix = os.environ.get("CONDA_PREFIX")
    conda_default_env = os.environ.get("CONDA_DEFAULT_ENV")

    # Define paths to (de)activate scripts
    activate_dir = os.path.join(conda_prefix, "etc/conda/activate.d")
    deactivate_dir = os.path.join(conda_prefix, "etc/conda/deactivate.d")
    print("Copying environment variable setup scripts to:\n")
    print(f"    {activate_dir}")
    print(f"    {deactivate_dir}\n")

    # Copy environment setup scripts
    os.makedirs(activate_dir, exist_ok=True)
    with open(os.path.join(activate_dir, "env_vars.sh"), "w") as f:
        f.write(activate_script(root_dir))

    os.makedirs(deactivate_dir, exist_ok=True)
    with open(os.path.join(deactivate_dir, "env_vars.sh"), "w") as f:
        f.write(deactivate_script())

    # fmt: off
    print(dedent(f"""
    You might want to reactivate your environment:

        $ conda deactivate && conda activate {conda_default_env}
    """).lstrip())
    # fmt: on


def activate_script(eradiate_source_dir):
    # fmt: off
    return dedent(f"""
        #!/bin/sh
        ERADIATE_SOURCE_DIR="{eradiate_source_dir}"
        source ${{ERADIATE_SOURCE_DIR}}/setpath.sh
    """).lstrip()
    # fmt: on


def deactivate_script():
    # fmt: off
    return dedent(f"""
        #!/bin/sh
        # Clean up PATH
        match="${{ERADIATE_SOURCE_DIR}}/ext/mitsuba/build:"
        export PATH=${{PATH//$match/}}
        # Clean up PYTHONPATH
        match="${{ERADIATE_SOURCE_DIR}}/ext/mitsuba/build/python:"
        export PYTHONPATH="${{PYTHONPATH//$match/}}"
        # Remove other environment variables
        unset ERADIATE_SOURCE_DIR
    """).lstrip()
    # fmt: on


if __name__ == "__main__":
    main()



import mitsuba as mi
import pytest

from eradiate.scenes.bsdfs import HapkeBSDF
from eradiate.test_tools.types import check_scene_element
from eradiate.units import unit_registry as ureg


@pytest.mark.parametrize(
    "kwargs",
    [
        {"w": 0.1, "b": 0.2, "c": 0.3, "theta": 40.0 * ureg.deg, "B_0": 0.5, "h": 0.6},
        {
            "w": 0.1,
            "b": 0.2,
            "c": {"type": "uniform", "value": 0.3},
            "theta": {
                "type": "interpolated",
                "wavelengths": [300.0, 800.0],
                "values": [10.0, 20.0],
            },
            "B_0": 0.5,
            "h": 0.6,
        },
    ],
    ids=["uniform", "mixed"],
)
def test_hapke_construct(modes_all_double, kwargs):
    bsdf = HapkeBSDF(**kwargs)

    assert bsdf
    check_scene_element(bsdf, mi.BSDF)



import joseki
import numpy as np
import pytest

from eradiate import data
from eradiate import unit_registry as ureg
from eradiate.radprops.absorption import (
    eval_sigma_a_ckd_impl,
    eval_sigma_a_mono_impl,
    wrange_ckd,
    wrange_mono,
)
from eradiate.units import to_quantity


@pytest.mark.parametrize(
    "w",
    [
        np.array([550.0]) * ureg.nm,
        np.linspace(540.0, 560.0) * ureg.nm,
    ],
)
def test_eval_sigma_a_mono_impl(w, error_handler_config):
    """
    The shape of the absorption coefficient array is consistent with the
    wavelength (w) and altitude (z) arrays.
    """
    z = np.linspace(0.0, 10.0, 11) * ureg.km
    thermoprops = joseki.make(
        identifier="afgl_1986-us_standard",
        z=z,
        additional_molecules=False,
    )
    ds = data.load_dataset("spectra/absorption/mono/komodo/komodo.nc")
    sigma_a = eval_sigma_a_mono_impl(
        absorption_data={wrange_mono(ds): ds},
        thermoprops=thermoprops,
        w=w,
        error_handler_config=error_handler_config,
    )

    # sigma_a should have a shape of (w, z)
    assert sigma_a.shape == (w.size, z.size)


def test_eval_sigma_a_ckd_impl(error_handler_config):
    """
    The shape of the absorption coefficient array is consistent with the
    wavelength (w) and altitude (z) arrays.
    """
    z = np.linspace(0.0, 10.0, 11) * ureg.km
    thermoprops = joseki.make(
        identifier="afgl_1986-us_standard",
        z=z,
        additional_molecules=False,
    )
    ds = data.load_dataset("spectra/absorption/ckd/monotropa/monotropa-18100_18200.nc")
    wcenter = to_quantity(ds.w)
    sigma_a = eval_sigma_a_ckd_impl(
        absorption_data={wrange_ckd(ds): ds},
        thermoprops=thermoprops,
        w=wcenter,
        g=0.5,
        error_handler_config=error_handler_config,
    )

    # sigma_a should have a shape of (w, z)
    assert sigma_a.shape == (1, z.size)



"""
Functions to compute atmospheric absorption properties.
"""
from __future__ import annotations

import logging
import warnings

import numpy as np
import pint
import xarray as xr

import eradiate

from ..exceptions import (
    InterpolationError,
    MissingCoordinateError,
    OutOfBoundsCoordinateError,
    ScalarCoordinateError,
    UnsupportedModeError,
)
from ..units import to_quantity
from ..units import unit_context_config as ucc
from ..units import unit_registry as ureg

logger = logging.getLogger(__name__)

# ------------------------------------------------------------------------------
#                       Interpolation errors handling
# ------------------------------------------------------------------------------

DEFAULT_X_ERROR_HANDLER_CONFIG = {
    "missing": "ignore",
    "scalar": "ignore",
    "bounds": "raise",
}

DEFAULT_P_ERROR_HANDLER_CONFIG = {
    "missing": "raise",
    "scalar": "raise",
    "bounds": "ignore",  # atmospheric pressure easily goes very low, but at
    # these altitude, the air is very thin and so is the
    # absorption coefficient
}

DEFAULT_T_ERROR_HANDLER_CONFIG = {
    "missing": "raise",
    "scalar": "raise",
    "bounds": "ignore",  # atmospheric temperature climbs fast above 80-100 km,
    # but at these altitude, the air is very thin and so is
    # the absorption coefficient
}

#: Default interpolation error handling configuration
DEFAULT_HANDLER_CONFIG = {
    "x": DEFAULT_X_ERROR_HANDLER_CONFIG,
    "p": DEFAULT_P_ERROR_HANDLER_CONFIG,
    "t": DEFAULT_T_ERROR_HANDLER_CONFIG,
}

ERROR_TYPE_TO_STR = {
    MissingCoordinateError: "missing",
    ScalarCoordinateError: "scalar",
    OutOfBoundsCoordinateError: "bounds",
}


def handle_error(error: InterpolationError, config):
    action = config[ERROR_TYPE_TO_STR[type(error)]]

    if action == "raise":
        raise error
    elif action == "warn":
        warnings.warn(str(error), UserWarning)
    elif action == "ignore":
        pass
    else:
        raise NotImplementedError(f"Action {action} not implemented.")


# ------------------------------------------------------------------------------
#                       Interpolation functions
# ------------------------------------------------------------------------------


def xcoords(k: xr.DataArray) -> list[str]:
    return [c for c in k.coords if c.startswith("x_")]


def xrange(k: xr.DataArray) -> dict[str, pint.Quantity]:
    _xrange = {}
    for c in xcoords(k):
        if k[c].size > 1:
            x = to_quantity(k[c])
            _xrange[c] = np.stack([x.min(), x.max()])
    return _xrange


def xcoords_scalar(k: xr.DataArray) -> list[str]:
    return [c for c in xcoords(k) if k[c].size == 1]


def prange(k: xr.DataArray) -> pint.Quantity:
    p = to_quantity(k["p"])
    return np.stack([p.min(), p.max()])


def trange(k: xr.DataArray) -> pint.Quantity:
    t = to_quantity(k["t"])
    return np.stack([t.min(), t.max()])


def check_missing_x_coords(x_ds: list[str], x_thermoprops: list[str]):
    if not set(x_thermoprops).issubset(set(x_ds)):
        msg = (
            "The absorption dataset does not contain all the volume fraction "
            "coordinates present in the thermophysical properties dataset."
        )
        msg += f"\nVolume fraction coordinates in absorption data set: {x_ds}"

        msg += (
            f"\nVolume fraction coordinates in thermophysical properties data "
            f"set: {x_thermoprops}"
        )
        msg += (
            f"\nMissing volume fraction coordinates: "
            f"{list(set(x_thermoprops) - set(x_ds))}"
        )
        raise MissingCoordinateError(msg)


def check_scalar_x_coords(
    x_absorption_scalar: list[str],
    thermoprops: xr.Dataset,
    rtol=1e-3,
):
    x_thermoprops = [f"x_{m}" for m in thermoprops.joseki.molecules]
    if set(x_absorption_scalar).issubset(set(x_thermoprops)):
        x_intersection = set(x_absorption_scalar).intersection(set(x_thermoprops))

        _x_nonconstant = []
        for _x in x_intersection:
            values = thermoprops[_x].values
            if not np.all(np.isclose(values, values[0], rtol=rtol)):
                _x_nonconstant.append(_x)

        if _x_nonconstant:
            msg = (
                f"These volume fraction coordinates show more than {rtol:.2%} "
                f"variation in the atmosphere's thermophysical properties "
                f"but are scalar coordinates in the absorption dataset: "
                f"{_x_nonconstant}"
            )
            raise ScalarCoordinateError(msg)


def check_x_ranges(k: xr.DataArray, thermoprops: xr.Dataset) -> None:
    xrange_ds = xrange(k)
    out_of_bounds = {}
    for c in xrange_ds:
        x_thermoprops = to_quantity(thermoprops[c])
        xrange_t = np.stack([x_thermoprops.min(), x_thermoprops.max()])
        if xrange_t[0] < xrange_ds[c][0] or xrange_t[1] > xrange_ds[c][1]:
            out_of_bounds[c] = {
                "thermoprops": xrange_t,
                "absorption": xrange_ds[c],
            }

    if out_of_bounds:
        msg = "The following volume fraction coordinates are out of bounds."
        for c in out_of_bounds:
            msg += f"\n{c}: range in thermoprops {out_of_bounds[c]['thermoprops']}, "
            msg += f"range in absorption dataset {out_of_bounds[c]['absorption']}"
        raise OutOfBoundsCoordinateError(msg)


def interp_x(
    k: xr.DataArray,
    thermoprops: xr.Dataset,
    error_handler_config: dict[str, str] | None = None,
) -> xr.DataArray:
    """
    Interpolate absorption coefficient data along volume fraction dimensions.

    Parameters
    ----------
    k : xr.DataArray
        Absorption coefficient data array.

    thermoprops : xr.Dataset
        Atmosphere's thermophysical properties.

    error_handler_config : dict[str, str], optional
        Configuration of the error handler, by default
        DEFAULT_X_ERROR_HANDLER_CONFIG

    Returns
    -------
    xr.DataArray
        Interpolated absorption coefficient data array.
    """
    # update missing keys in handle_config with DEFAULT_HANDLE_CONFIG
    if error_handler_config is None:
        error_handler_config = DEFAULT_X_ERROR_HANDLER_CONFIG
    else:
        error_handler_config = {
            **DEFAULT_X_ERROR_HANDLER_CONFIG,
            **error_handler_config,
        }

    x_thermoprops = [f"x_{m}" for m in thermoprops.joseki.molecules]
    logger.debug("x_thermoprops: %s", x_thermoprops)

    x_absorption = xcoords(k)
    x_absorption_scalar = xcoords_scalar(k)
    x_absorption_array = set(x_absorption) - set(x_absorption_scalar)
    x_absorption_interp = x_absorption_array  # interpolation along these coords
    x_absorption_sel = x_absorption_scalar  # selection along these coords
    logger.debug("x_absorption_interp: %s", x_absorption_interp)
    logger.debug("x_absorption_sel: %s", x_absorption_sel)

    # -------------------------------------------------------------------------
    #               Check coordinates for interpolation errors
    # -------------------------------------------------------------------------

    # Unspecified coordinates (always raise)  # TODO: or set this coord to zero?
    # a coordinate is present in 'k' but not in 'thermoprops'
    if not set(x_absorption).issubset(set(x_thermoprops)):
        msg = (
            f"The absorption dataset contains array coordinates ("
            f"{list(set(x_absorption) - set(x_thermoprops))} that are "
            f"not specified in the thermophysical properties dataset."
        )
        logger.critical(msg)
        raise ValueError(msg)

    # Missing coordinates
    # a coordinate is specified in thermoprops but not in 'k'
    try:
        check_missing_x_coords(x_absorption, x_thermoprops)
    except MissingCoordinateError as e:
        handle_error(e, error_handler_config)

    # Scalar coordinates
    # a coordinate is present (and varying) in thermoprops that corresponds to
    # a scalar (one-value) coordinate in 'k'
    try:
        check_scalar_x_coords(x_absorption_scalar, thermoprops)
    except ScalarCoordinateError as e:
        handle_error(e, error_handler_config)

    # Out of bounds coordinate
    # a coordinate value in thermoprops is out of the range of the
    # corresponding coordinate in 'k'
    try:
        check_x_ranges(k, thermoprops)
        bounds_error = True
    except OutOfBoundsCoordinateError as e:
        handle_error(e, error_handler_config)
        bounds_error = False

    # -------------------------------------------------------------------------
    #                   Select and/or interpolate
    # -------------------------------------------------------------------------

    # Select
    k_selected = k.isel(**{x: 0 for x in x_absorption_sel}, drop=True)

    # Interpolate
    fill_value = None if bounds_error else 0.0  # TODO: use 2-element tuple?

    k_interp = k_selected.interp(
        **{x: thermoprops[x] for x in x_absorption_interp},
        kwargs={
            "bounds_error": bounds_error,
            "fill_value": fill_value,
        },
    ).drop_vars(x_absorption_interp)

    return k_interp


def check_p_range(k: xr.Dataset, thermoprops: xr.Dataset):
    p_thermoprops = to_quantity(thermoprops.p)
    p_thermoprops_range = np.stack([p_thermoprops.min(), p_thermoprops.max()])
    p_k = to_quantity(k.p)
    p_k_range = np.stack([p_k.min(), p_k.max()])

    if p_thermoprops.min() < p_k.min() or p_thermoprops.max() > p_k.max():
        raise OutOfBoundsCoordinateError(
            f"The pressure range of the thermophysical properties dataset "
            f"({p_thermoprops_range}) is outside the pressure range of the "
            f"absorption dataset ({p_k_range})."
        )


def interp_p(
    k: xr.DataArray,
    thermoprops: xr.Dataset,
    error_handler_config: dict[str, str] | None = None,
) -> xr.Dataset:
    """
    Interpolate absorption dataset along pressure dimension.

    Parameters
    ----------
    k : xr.DataArray
        Absorption coefficient data array.

    thermoprops : xr.Dataset
        Atmosphere's thermophysical properties.

    error_handler_config : dict[str, str], optional
        Configuration of the error handler, by default
        DEFAULT_X_ERROR_HANDLER_CONFIG

    Returns
    -------
    xr.DataArray
        Interpolated absorption coefficient data array.
    """
    try:
        check_p_range(k, thermoprops)
        bounds_error = True
    except OutOfBoundsCoordinateError as e:
        if error_handler_config is None:
            error_handler_config = DEFAULT_P_ERROR_HANDLER_CONFIG
        else:
            error_handler_config = {
                **DEFAULT_P_ERROR_HANDLER_CONFIG,
                **error_handler_config,
            }
        handle_error(e, error_handler_config)
        bounds_error = False

    fill_value = None if bounds_error else 0.0  # TODO: use 2-element tuple?

    k_interp = k.interp(
        p=thermoprops.p,
        kwargs={"bounds_error": bounds_error, "fill_value": fill_value},
    ).drop_vars("p")

    return k_interp


def check_t_range(k: xr.DataArray, thermoprops: xr.Dataset):
    t_thermoprops = to_quantity(thermoprops.t)
    t_thermoprops_range = np.stack([t_thermoprops.min(), t_thermoprops.max()])
    t_ds = to_quantity(k.t)
    t_ds_range = np.stack([t_ds.min(), t_ds.max()])

    if t_thermoprops.min() < t_ds.min() or t_thermoprops.max() > t_ds.max():
        raise OutOfBoundsCoordinateError(
            f"The temperature range of the thermophysical properties dataset "
            f"({t_thermoprops_range}) is outside the temperature range of the "
            f"absorption dataset ({t_ds_range})."
        )


def interp_t(
    k: xr.DataArray,
    thermoprops: xr.Dataset,
    error_handler_config: dict[str, str] | None = None,
) -> xr.Dataset:
    """
    Interpolate absorption coefficient data set along temperature dimension.

    Parameters
    ----------
    k : xr.DataArray
        Absorption coefficient data array.

    thermoprops : xr.Dataset
        Atmosphere's thermophysical properties.

    error_handler_config : dict[str, str], optional
        Configuration of the error handler, by default
        DEFAULT_X_ERROR_HANDLER_CONFIG

    Returns
    -------
    xr.DataArray
        Interpolated absorption coefficient data array.
    """
    try:
        check_t_range(k, thermoprops)
        bounds_error = True
    except OutOfBoundsCoordinateError as e:
        if error_handler_config is None:
            error_handler_config = DEFAULT_T_ERROR_HANDLER_CONFIG
        else:
            error_handler_config = {
                **DEFAULT_T_ERROR_HANDLER_CONFIG,
                **error_handler_config,
            }
        handle_error(e, error_handler_config)
        bounds_error = False

    fill_value = None if bounds_error else 0.0  # TODO: use 2-element tuple?

    k_interp = k.interp(
        t=thermoprops.t,
        kwargs={"bounds_error": bounds_error, "fill_value": fill_value},
    ).drop_vars("t")

    return k_interp


# ------------------------------------------------------------------------------
#       Absorption coefficient evaluation implementation in mono modes
# ------------------------------------------------------------------------------


def wrange_mono(ds: xr.Dataset) -> tuple[pint.Quantity]:
    wds = to_quantity(ds.w)
    wunits = ucc.get("wavelength")

    if wds.check("[length]^-1"):
        return ((1.0 / wds.max()).to(wunits), (1.0 / wds.min()).to(wunits))
    elif wds.check("[length]"):
        return (wds.min().to(wunits), wds.max().to(wunits))
    else:
        raise ValueError(
            f"Spectral coordinate of absorption dataset has unexpected units "
            f"({ds.w.units})."
        )


def check_w_range_mono(ds: xr.Dataset, w: pint.Quantity):
    wmin, wmax = wrange_mono(ds=ds)

    if np.any((w < wmin) | (w > wmax)):
        msg = (
            f"Requested w coordinate ({w}) is outside the range of the "
            f"absorption data set ({wmin} to {wmax})."
        )
        raise OutOfBoundsCoordinateError(msg)


def interp_w(ds: xr.Dataset, w: pint.Quantity) -> xr.DataArray:
    """
    Interpolate absorption data set to a given wavelength.

    Parameters
    ----------
    ds : xr.Dataset
        Absorption data set.

    w : pint.Quantity
        Wavelength.

    Returns
    -------
    xr.DataArray
        Wavelength-interpolated absorption data.
    """
    # convert to wavenumber/wavelength
    wunits = ds.w.units
    if ureg(wunits).check("[length]^-1"):
        # interpolation according to wavenumber
        wm = (1 / w).m_as(wunits)
    elif ureg(wunits).check("[length]"):
        # interpolation according to wavelength
        wm = w.m_as(wunits)
    else:
        raise ValueError(f"Cannot interpret units {wunits}")

    # check wavelength range and interpolate
    check_w_range_mono(ds, w)
    return ds.sigma_a.interp(
        w=wm,
        method="linear",
    )


def eval_sigma_a_mono_impl_ds(
    ds: xr.Dataset,
    thermoprops: xr.Dataset,
    w: pint.Quantity,
    error_handler_config: dict[str, dict[str, str]] | None = None,
) -> pint.Quantity:
    """
    Evaluate absorption coefficient in given spectral and thermophysical
    conditions.

    Parameters
    ----------
    ds : Dataset
        Absorption coefficient dataset.

    thermoprops : Dataset
        Atmospheric thermophysical properties data set.

    w : quantity
        Wavelength.

    error_handler_config : dict
        Error handler configuration.

    Returns
    -------
    quantity
        Absorption coefficient values.
    """
    if error_handler_config is None:
        error_handler_config = DEFAULT_HANDLER_CONFIG

    kw = interp_w(ds, w)
    kwx = interp_x(kw, thermoprops, error_handler_config["x"])
    kwxp = interp_p(kwx, thermoprops, error_handler_config["p"])
    kwxpt = interp_t(kwxp, thermoprops, error_handler_config["t"])
    return kwxpt


def eval_sigma_a_mono_impl(
    absorption_data: dict[tuple[pint.Quantity], xr.Dataset],
    thermoprops: xr.Dataset,
    w: pint.Quantity,
    error_handler_config: dict[str, dict[str, str]] | None = None,
) -> pint.Quantity:
    # NOTE: it is assumed that wavelength intervals do no intersect, namely that there
    # is at most one dataset relevant to compute the absorption coefficient at
    # a given wavelength
    # TODO: check that in absorption_data validator

    # NOTE: if the evaluation wavelengths span only one dataset, the evaluation
    # is vectorized. Otherwise, the evaluation is vectorized only when each
    # individual dataset is evaluated.

    if error_handler_config is None:
        error_handler_config = DEFAULT_HANDLER_CONFIG

    w = np.atleast_1d(w)
    wargsort = np.argsort(w)
    ws = w[wargsort]
    wargunsort = np.argsort(wargsort)
    # assume that w might be a large array

    das = []
    upper_max = max([i[1] for i in absorption_data.keys()])
    for interval, dataset in absorption_data.items():
        # we assume that this list is not too long (<10 elements) so the
        # performance cost of this loop is not too high.

        # find the wavelength that are contained in the current wavelength
        # interval:
        if interval[1] != upper_max:
            w_where = ws[
                (ws >= interval[0]) & (ws < interval[1])  # mind the strict inequality
            ]
        else:
            w_where = ws[(ws >= interval[0]) & (ws <= interval[1])]

        if w_where.size > 0:
            da = eval_sigma_a_mono_impl_ds(
                ds=dataset,
                thermoprops=thermoprops,
                w=w_where,
                error_handler_config=error_handler_config,
            )
            das.append(da)

    concatenated = xr.concat(das, dim="w")
    return to_quantity(concatenated.isel(w=wargunsort).transpose("w", "z"))


# ------------------------------------------------------------------------------
#       Absorption coefficient evaluation implementation in CKD modes
# ------------------------------------------------------------------------------


def wrange_ckd(ds: xr.Dataset) -> tuple[pint.Quantity]:
    wbounds = to_quantity(ds.wbounds.squeeze())
    wunits = ucc.get("wavelength")
    if wbounds.check("[length]^-1"):
        return tuple(np.sort((1 / wbounds.to(wunits))))
    elif wbounds.check("[length]"):
        return tuple(np.sort(wbounds).to(wunits))
    else:
        raise ValueError(
            f"Spectral coordinate of absorption dataset has unexpected units "
            f"({ds.w.units})."
        )


def check_w_range_ckd(ds: xr.Dataset, w: pint.Quantity, rtol: float = 1e-3):
    # This function is both for CKD absorption datasets

    # Check that the evaluation wavelength is amongst the absorption dataset
    # spectral coordinate values (to a relative tolerance of 'rtol').

    wds = to_quantity(ds.w)
    if wds.check("[length]^-1"):
        w_k = (1 / wds).to(w.units)
    elif wds.check("[length]"):
        w_k = wds
    else:
        raise ValueError(
            f"Spectral coordinate of absorption dataset has unexpected units "
            f"({ds.w.units})."
        )

    if not np.any(np.isclose(w_k, w, rtol=rtol)):
        msg = (
            f"Requested w coordinate ({w}) is not amongst the spectral "
            f"coordinates of the absorption data set ({w_k})."
        )
        raise OutOfBoundsCoordinateError(msg)


def interp_wg(ds: xr.Dataset, w: pint.Quantity, g: float) -> xr.DataArray:
    """
    Interpolate absorption data set to given w and g coordinates.

    Parameters
    ----------
    ds : xr.Dataset
        Absorption data set.

    w : pint.Quantity
        Wavelength corresponding to the CKD bin to select.

    g : float
        g-point to interpolate at.

    Returns
    -------
    xr.DataArray
        Spectral-interpolated absorption data.
    """
    # 1. bin selection
    wunits = ds.w.units
    if ureg(wunits).check("[length]^-1"):
        # select bin according to wavenumber
        wm = (1 / w).m_as(wunits)
    elif ureg(wunits).check("[length]"):
        # select bin according to wavelength
        wm = w.m_as(wunits)
    else:
        raise ValueError(f"Cannot interpret units {wunits}")

    check_w_range_ckd(ds, w, rtol=1e-3)
    # select the bin
    kw = ds.sigma_a.sel(
        w=wm,
        method="nearest",
    ).expand_dims("w")

    # 2. interpolation along g-point (pseudo-spectral coordinate)
    return kw.interp(g=g).drop_vars("g")


def eval_sigma_a_ckd_impl_ds(
    ds: xr.Dataset,
    thermoprops: xr.Dataset,
    w: pint.Quantity,
    g: float,
    error_handler_config: dict[str, dict[str, str]] | None = None,
) -> pint.Quantity:
    """
    Evaluate absorption coefficient in given spectral and thermophysical
    conditions.

    Parameters
    ----------
    ds : Dataset
        Absorption coefficient dataset.

    thermoprops : Dataset
        Atmospheric thermophysical properties data set.

    w : quantity
        Wavelength.

    g: float
        g-point.

    error_handler_config : dict
        Error handler configuration.

    Returns
    -------
    quantity
        Absorption coefficient values.
    """
    if error_handler_config is None:
        error_handler_config = DEFAULT_HANDLER_CONFIG

    kw = interp_wg(ds, w, g)
    kwx = interp_x(kw, thermoprops, error_handler_config["x"])
    kwxp = interp_p(kwx, thermoprops, error_handler_config["p"])
    kwxpt = interp_t(kwxp, thermoprops, error_handler_config["t"])
    return kwxpt


def _isclose(a, b, **kwargs):
    """
    Return values in b that are close to one value in a (different shapes supported).
    """
    out = []
    for value in a:
        isclosed = b[np.isclose(b, value, **kwargs)]
        if len(isclosed) > 0:
            out.append(isclosed)
    if len(out) > 0:
        return np.stack(out).squeeze()
    else:
        return np.empty(0)


def w_dataset(ds: xr.Dataset) -> pint.Quantity:
    w = to_quantity(ds.w)
    wunits = ucc.get("wavelength")
    if w.check("[length]"):
        return w.to(wunits)
    elif w.check("[length^-1]"):
        return (1.0 / w).to(wunits)
    else:
        raise ValueError


def eval_sigma_a_ckd_impl(
    absorption_data,
    thermoprops: xr.Dataset,
    w: pint.Quantity,
    g: float,
    error_handler_config: dict[str, dict[str, str]] | None = None,
) -> pint.Quantity:
    # NOTE: it is assumed that only one dataset corresponds to each wavelength
    # value in 'w'

    if error_handler_config is None:
        error_handler_config = DEFAULT_HANDLER_CONFIG

    w = np.atleast_1d(w)
    wargsort = np.argsort(w)
    ws = w[wargsort]
    wargunsort = np.argsort(wargsort)

    das = []
    for _, dataset in absorption_data.items():
        # we assume that this list is not too long (~10 elements) so the
        # performance cost of this loop is not too high.

        # find the wavelength values that match with the current dataset
        w_ds = w_dataset(dataset)
        w_where = _isclose(ws, w_ds, rtol=1e-3)

        if w_where.size > 0:
            da = eval_sigma_a_ckd_impl_ds(
                ds=dataset,
                thermoprops=thermoprops,
                w=w_where,
                g=g,
                error_handler_config=error_handler_config,
            )
            das.append(da)

    if len(das) == 0:
        w_dss = []
        w_wheres = []
        for _, dataset in absorption_data.items():
            w_ds = w_dataset(dataset)
            w_where = _isclose(ws, w_ds, rtol=1e-3)
            w_dss.append(w_ds)
            w_wheres.append(w_where)

        raise ValueError(f"{w_dss=}, {w_wheres=}")

    concatenated = xr.concat(das, dim="w")
    return to_quantity(concatenated.isel(w=wargunsort).transpose("w", "z"))


# ------------------------------------------------------------------------------
#                wavelength range based on active mode
# ------------------------------------------------------------------------------


def wrange(ds: xr.Dataset):
    if eradiate.mode().is_mono:
        wrange = wrange_mono(ds)
    elif eradiate.mode().is_ckd:
        wrange = wrange_ckd(ds)
    else:
        raise UnsupportedModeError
    return wrange



"""
This module tests the post-processing pipeline logic. It uses parametrized
fixtures to make the process of generating the same data multiple times as easy
and comprehensive as possible. The processing part of testing is implemented by
fixtures; test bodies perform only data checks.
"""

from __future__ import annotations

import numpy as np
import pytest
import xarray as xr

import eradiate
import eradiate.pipelines.logic as logic
from eradiate.experiments import AtmosphereExperiment
from eradiate.scenes.illumination import ConstantIllumination, DirectionalIllumination
from eradiate.spectral import BinSet, WavelengthSet
from eradiate.units import unit_registry as ureg

# ------------------------------------------------------------------------------
#                                    Fixtures
# ------------------------------------------------------------------------------

# In order to ensure that each test step gets the data formatted as it expects,
# we provide here a fixture-based implementation of a "mini-pipeline" that
# chains the operations that are unit-tested.


@pytest.fixture(params=["mono", "ckd"])
def mode_id(request):
    # This fixture parametrizes the `experiment` fixture against the selected
    # mode
    return request.param


@pytest.fixture(params=["hemispherical_distant", "distant_flux"])
def measure(request):
    # This fixture is used to further parametrize the `experiment` fixture
    if request.param in {"hemispherical_distant", "distant_flux"}:
        return request.param

    raise NotImplementedError(request.param)


@pytest.fixture(params=["multi_delta", "sentinel_2a-msi-3"])
def srf(request):
    # This fixture is used to further parametrize the `experiment` fixture
    if request.param in {"multi_delta", "sentinel_2a-msi-3"}:
        return request.param

    raise NotImplementedError(request.param)


@pytest.fixture
def mode(mode_id):
    # In addition to setting the mode, this fixture returns it for convenience
    eradiate.set_mode(mode_id)
    return eradiate.mode()


@pytest.fixture
def experiment(mode, srf, measure):
    # Create an experiment and run it (parametrized by above fixtures)

    if srf == "multi_delta":
        srf_dict = {"type": "multi_delta", "wavelengths": 550.0 * ureg.nm}
    elif srf == "sentinel_2a-msi-3":
        srf_dict = "sentinel_2a-msi-3"
    else:
        raise NotImplementedError(srf)

    exp = AtmosphereExperiment(
        atmosphere=None,
        surface={"type": "lambertian", "reflectance": 1.0},
        illumination={"type": "directional", "irradiance": 2.0},
        measures={
            "type": measure,
            "film_resolution": (32, 32),
            "spp": 256,
            "srf": srf_dict,
        },
    )
    exp.process()
    return exp


@pytest.fixture
def irradiance(mode_id, experiment):
    return logic.extract_irradiance(
        mode_id, experiment.illumination, experiment.spectral_set[0]
    )["irradiance"]


@pytest.fixture
def solar_angles(mode_id, experiment):
    return logic.extract_irradiance(
        mode_id, experiment.illumination, experiment.spectral_set[0]
    )["solar_angles"]


@pytest.fixture
def viewing_angles(experiment):
    return logic.viewing_angles(experiment.measures[0].viewing_angles.m_as("deg"))


@pytest.fixture
def raw_results(experiment):
    # Convenience: Return the raw results associated to an experiment
    return experiment.measures[0].mi_results


@pytest.fixture
def gather_bitmaps(mode, experiment, raw_results, viewing_angles, solar_angles):
    # Apply the `gather_bitmaps` pipeline step and return the results
    var_name, var_metadata = experiment.measures[0].var
    return logic.gather_bitmaps(
        mode.id, var_name, var_metadata, raw_results, viewing_angles, solar_angles
    )


@pytest.fixture
def aggregate_ckd_quad(mode, experiment, gather_bitmaps):
    var_name = experiment.measures[0].var[0]
    results_raw = gather_bitmaps[f"{var_name}_raw"]
    return logic.aggregate_ckd_quad(mode.id, results_raw, experiment.spectral_set[0])


@pytest.fixture
def apply_spectral_response(mode, measure, experiment, aggregate_ckd_quad):
    spectral_data = aggregate_ckd_quad
    return logic.apply_spectral_response(spectral_data, experiment.measures[0].srf)


# ------------------------------------------------------------------------------
#                                     Tests
# ------------------------------------------------------------------------------


@pytest.mark.parametrize("mode_id", ["ckd"])
@pytest.mark.parametrize("measure", ["hemispherical_distant"])
def test_aggregate_ckd_quad(experiment, gather_bitmaps, aggregate_ckd_quad):
    var_name = experiment.measures[0].var[0]
    raw = gather_bitmaps[f"{var_name}_raw"]
    result = aggregate_ckd_quad

    # Dimension checks
    expected_dims = set(raw.dims) - {"g"}
    assert set(result.dims) == expected_dims

    # Coordinate checks
    expected_coords = (set(raw.coords) - {"g"}) | {"bin_wmin", "bin_wmax"}
    assert set(result.coords) == expected_coords
    assert result.bin_wmin.dims == ("w",)
    assert result.bin_wmax.dims == ("w",)

    # Variable checks
    # -- In the present case, the quadrature evaluates to 2/π
    assert np.allclose(result.values, 2.0 / np.pi)
    # -- Metadata of the variable for which aggregation is performed are copied
    assert result.attrs == raw.attrs


@pytest.mark.parametrize("mode_id", ["ckd"])
@pytest.mark.parametrize("srf", ["sentinel_2a-msi-3"])
def test_apply_spectral_response_main(
    experiment, aggregate_ckd_quad, apply_spectral_response
):
    """
    Unit test for :func:`.apply_spectral_response` when used on the main
    variable.
    """
    var_name = experiment.measures[0].var[0]
    raw = aggregate_ckd_quad
    result = apply_spectral_response

    # Dimension checks
    expected_dims = set(raw.dims) - {"w"}
    assert set(result.dims) == expected_dims

    # Coordinate checks
    expected_coords = set(raw.coords) - ({"w"} | {"bin_wmax", "bin_wmin"})
    assert set(result.coords) == expected_coords

    # The step adds an SRF-weighted variable
    assert apply_spectral_response.name == f"{var_name}_srf"
    assert np.all(apply_spectral_response.values > 0.0)


@pytest.mark.parametrize("mode_id", ["ckd"])
@pytest.mark.parametrize("srf", ["sentinel_2a-msi-3"])
def test_apply_spectral_response_irradiance(irradiance, experiment):
    """
    Unit test for :func:`.apply_spectral_response` when used on the irradiance
    variable.
    """
    result = logic.apply_spectral_response(irradiance, experiment.measures[0].srf)

    # The step adds an SRF-weighted variable
    assert result.name == "irradiance_srf"
    assert np.all(result.values > 0.0)


@pytest.mark.parametrize(
    "illumination_type, expected_dims, expect_solar_angles",
    [
        ("directional", ["sza", "saa"], True),
        ("constant", [], False),
    ],
)
def test_extract_irradiance(
    mode, illumination_type, expected_dims, expect_solar_angles
):
    if illumination_type == "directional":
        illumination = DirectionalIllumination(zenith=30.0, azimuth=45.0)
    elif illumination_type == "constant":
        illumination = ConstantIllumination()
    else:
        raise ValueError

    # Computation succeeds
    if mode.is_mono:
        spectral_set = WavelengthSet(np.linspace(400.0, 500.0) * ureg.nm)
    elif mode.is_ckd:
        spectral_set = BinSet.arange(400.0 * ureg.nm, 500.0 * ureg.nm, 10.0 * ureg.nm)
    else:
        raise NotImplementedError

    # Return value is a dictionary holding a data array and an optional dataset
    result = logic.extract_irradiance(mode.id, illumination, spectral_set)
    assert set(result.keys()) == {"irradiance", "solar_angles"}
    assert isinstance(result["irradiance"], xr.DataArray)

    if expect_solar_angles:
        assert isinstance(result["solar_angles"], xr.Dataset)
    else:
        assert result["solar_angles"] is None

    # Irradiance is indexed by solar angle and spectral coordinates
    assert set(result["irradiance"].dims) == {"w"}.union(set(expected_dims))

    # Irradiance data array also contains bin bounds as coordinates when relevant
    expected_coords = {"w"}.union(set(expected_dims))
    if eradiate.mode().is_ckd:
        expected_coords |= {"bin_wmin", "bin_wmax"}
    assert set(result["irradiance"].coords) == expected_coords


@pytest.mark.parametrize("srf", ["multi_delta"])
@pytest.mark.parametrize("measure", ["hemispherical_distant"])
def test_gather_bitmaps(mode, gather_bitmaps):
    # Routine creates the variables we expect
    expected_variables = {"spp", "radiance_raw", "weights_raw"}
    assert set(gather_bitmaps.keys()) == expected_variables

    # Each variable has the dimensions we expect
    # Note: the 'channel' film dimension is dropped in mono mode
    if mode.is_mono:
        spectral_sizes = {"w": 1}
    elif mode.is_ckd:
        spectral_sizes = {"w": 1, "g": 1}  # 1 is the default value
    else:
        raise RuntimeError("Selected mode is not handled")

    solar_angle_sizes = {"sza": 1, "saa": 1}
    film_sizes = {"y_index": 32, "x_index": 32}
    all_sizes = {**spectral_sizes, **film_sizes, **solar_angle_sizes}

    expected_sizes = {"spp": spectral_sizes, "radiance_raw": all_sizes}
    for var, da in gather_bitmaps.items():  # noqa: F402
        if da is None:
            continue
        assert da.sizes == expected_sizes[var], (
            f"While checking variable '{var}', expected sizes '{expected_sizes[var]}', "
            f"got '{da.sizes}'"
        )

    # Check radiance values
    assert np.allclose(gather_bitmaps["radiance_raw"].values, 2.0 / np.pi)


@pytest.mark.parametrize("measure", ["distant_flux"])
def test_radiosity(mode, gather_bitmaps):
    # Initialize test data
    irradiance = 2.0
    sector_radiosity = gather_bitmaps["sector_radiosity_raw"]

    # Configure and apply step
    result = logic.radiosity(sector_radiosity=sector_radiosity)

    # Check that radiosity dimensions are correct
    assert not {"x_index", "y_index"}.issubset(result.dims)
    # This setup conserves energy
    assert np.allclose(irradiance, result, rtol=1e-3)


@pytest.mark.parametrize("mode_id", ["mono"])
@pytest.mark.parametrize("srf", ["multi_delta"])
def test_viewing_angles(experiment):
    # This test is minimal and simply checks for returned type and variable names
    result = logic.viewing_angles(experiment.measures[0].viewing_angles)
    assert isinstance(result, xr.Dataset)
    assert set(result.data_vars) == {"vaa", "vza"}



from __future__ import annotations

import attrs
import drjit as dr
import mitsuba as mi
import numpy as np
import pint
import pinttr

from ._core import Illumination
from ..core import NodeSceneElement
from ..spectra import SolarIrradianceSpectrum, Spectrum, spectrum_factory
from ..._config import config
from ...attrs import documented, parse_docs
from ...frame import AzimuthConvention, angles_to_direction
from ...units import unit_context_config as ucc
from ...units import unit_registry as ureg
from ...validators import has_quantity, is_positive


@parse_docs
@attrs.define(eq=False, slots=False)
class DirectionalIllumination(Illumination):
    """
    Directional illumination scene element [``directional``].

    The illumination is oriented based on the classical angular convention used
    in Earth observation.
    """

    zenith: pint.Quantity = documented(
        pinttr.field(
            default=0.0 * ureg.deg,
            validator=[is_positive, pinttr.validators.has_compatible_units],
            units=ucc.deferred("angle"),
        ),
        doc="Zenith angle.\n\nUnit-enabled field (default units: ucc[angle]).",
        type="quantity",
        init_type="quantity or float",
        default="0.0 deg",
    )

    azimuth: pint.Quantity = documented(
        pinttr.field(
            default=0.0 * ureg.deg,
            validator=[is_positive, pinttr.validators.has_compatible_units],
            units=ucc.deferred("angle"),
        ),
        doc="Azimuth angle value.\n"
        "\n"
        "Unit-enabled field (default units: ucc[angle]).",
        type="quantity",
        init_type="quantity or float",
        default="0.0 deg",
    )

    azimuth_convention: AzimuthConvention = documented(
        attrs.field(
            default=None,
            converter=lambda x: config.azimuth_convention
            if x is None
            else (AzimuthConvention[x.upper()] if isinstance(x, str) else x),
            validator=attrs.validators.instance_of(AzimuthConvention),
        ),
        doc="Azimuth convention. If ``None``, the global default configuration "
        "is used (see :class:`.EradiateConfig`).",
        type=".AzimuthConvention",
        init_type=".AzimuthConvention or str, optional",
        default="None",
    )

    irradiance: Spectrum = documented(
        attrs.field(
            factory=SolarIrradianceSpectrum,
            converter=spectrum_factory.converter("irradiance"),
            validator=[
                attrs.validators.instance_of(Spectrum),
                has_quantity("irradiance"),
            ],
        ),
        doc="Emitted power flux in the plane orthogonal to the illumination direction. "
        "Must be an irradiance spectrum (in W/m²/nm or compatible unit). "
        "Can be initialised with a dictionary processed by "
        ":meth:`.SpectrumFactory.convert`.",
        type=":class:`~eradiate.scenes.spectra.Spectrum`",
        init_type=":class:`~eradiate.scenes.spectra.Spectrum` or dict or float",
        default=":class:`SolarIrradianceSpectrum() <.SolarIrradianceSpectrum>`",
    )

    @property
    def direction(self) -> np.ndarray:
        """
        Illumination direction as an array of shape (3,), pointing inwards.
        """
        return angles_to_direction(
            [self.zenith.m_as(ureg.rad), self.azimuth.m_as(ureg.rad)],
            azimuth_convention=self.azimuth_convention,
            flip=True,
        ).reshape((3,))

    @property
    def _to_world(self) -> mi.ScalarTransorm4f:
        direction = dr.normalize(mi.ScalarVector3f(self.direction))
        up, _ = mi.coordinate_system(direction)
        return mi.ScalarTransform4f.look_at(origin=0.0, target=direction, up=up)

    @property
    def template(self) -> dict:
        return {"type": "directional", "to_world": self._to_world}

    @property
    def objects(self) -> dict[str, NodeSceneElement]:
        return {"irradiance": self.irradiance}



import mitsuba as mi
import numpy as np

import eradiate
from eradiate import KernelContext
from eradiate.kernel import mi_render, mi_traverse


def test_maximum_scene_size(modes_all_mono, json_metadata):
    r"""
    Maximum scene size (``path``)
    =============================

    This test searches for an order of magnitude of the maximum size a scene using
    a ``hdistant`` sensor without ray origin control can have. An arbitrary
    threshold is used as the pass/fail criterion for regression control.

    Rationale
    ---------

    - Geometry: a square surface with increasing sizes from 1.0 to 1e9.
      and a Lambertian BRDF with reflectance :math:`\rho = 0.5`.
    - Illumination: a directional light source at the zenith with radiance
      :math:`L_\mathrm{i} = 1.0`.
    - Sensor: a ``hdistant`` sensor targeting (0, 0, 0).

    Expected behaviour
    ------------------

    For all scene sizes below the parametrized size :code:`expected_min_size`
    the computational results must be equal to the theoretical prediction.
    Tolerance is set according to the defaults for :func:`numpy.allclose`.
    Metrics are reported only for the double precision version of this test.
    """
    expected_min_size = 1e8 if eradiate.mode().is_single_precision else 1e12
    spp = 1
    rho = 0.5
    li = 1.0
    expected = rho * li / np.pi
    max_i = 12
    scene_sizes = np.array(
        sorted(
            [10.0**i for i in range(1, max_i)]
            + [2.0 * 10**i for i in range(1, max_i)]
            + [5.0 * 10**i for i in range(1, max_i)]
            + [10.0**max_i]
        )
    )
    passed = np.full_like(scene_sizes, False, dtype=bool)

    for i, scene_size in enumerate(scene_sizes):
        mi_wrapper = mi_traverse(
            mi.load_dict(
                {
                    "type": "scene",
                    "bsdf_surface": {
                        "type": "diffuse",
                        "reflectance": rho,
                    },
                    "surface": {
                        "type": "rectangle",
                        "to_world": mi.ScalarTransform4f.scale(
                            mi.ScalarVector3f(scene_size, scene_size, 1)
                        ),
                        "bsdf": {"type": "ref", "id": "bsdf_surface"},
                    },
                    "illumination": {
                        "type": "directional",
                        "direction": [0, 0, -1],
                        "irradiance": li,
                    },
                    "measure": {
                        "type": "hdistant",
                        "id": "measure",
                        "target": [0, 0, 0],
                        "sampler": {"type": "independent", "sample_count": spp},
                        "film": {
                            "type": "hdrfilm",
                            "width": 32,
                            "height": 32,
                            "pixel_format": "luminance",
                            "component_format": "float32",
                            "rfilter": {"type": "box"},
                        },
                    },
                    "integrator": {"type": "path"},
                }
            )
        )

        result = np.squeeze(
            mi_render(mi_wrapper, ctxs=[KernelContext()])[550.0]["measure"]
        )
        passed[i] = np.allclose(result, expected)

    # Report test metrics
    max_size = float(np.max(scene_sizes[passed])) if np.any(passed) else 0.0

    if eradiate.mode().is_double_precision:
        json_metadata["metrics"] = {
            "test_maximum_scene_size": {
                "name": "Maximum scene size",
                "description": "The maximum scene size is: ",
                "value": f"{max_size:1.1e}",
                "units": "length units",
            }
        }

    # Maximum size for which result is correct is greater than requested min size
    assert max_size >= expected_min_size
    # All sizes below the threshold produce correct results
    assert np.all(passed[scene_sizes <= expected_min_size])



import logging

import mitsuba as mi
import pytest

import eradiate


def test_logging_setup(mode_mono):
    eradiate.kernel.install_logging()
    assert eradiate.kernel.logging.mi_logger is eradiate.kernel.logging._get_logger()


def test_logging_mitsuba(mode_mono, caplog):
    eradiate.kernel.install_logging()

    # TRACE log level is defined and set to the expected value
    assert logging.TRACE == logging.DEBUG - 5

    # Set log capture fixture level such that we can capture all log output
    caplog.set_level(logging.TRACE)

    # Trace and debug are not shown by default (Mitsuba kills them before they
    # can be passed to the Python logger)
    mi.Log(mi.LogLevel.Trace, "trace message")
    mi.Log(mi.LogLevel.Debug, "debug message")
    assert not caplog.records

    # We can raise log level
    eradiate.kernel.logging.mi_logger.set_log_level(mi.LogLevel.Trace)
    mi.Log(mi.LogLevel.Trace, "trace message")
    record = caplog.records.pop()
    assert record.levelname == "TRACE"
    assert "trace message" in record.msg

    mi.Log(mi.LogLevel.Debug, "debug message")
    record = caplog.records.pop()
    assert record.levelname == "DEBUG"
    assert "debug message" in record.msg

    # Info message is displayed
    mi.Log(mi.LogLevel.Info, "info message")
    record = caplog.records.pop()
    assert record.levelname == "INFO"
    assert "info message" in record.msg

    # Warning message as well
    mi.Log(mi.LogLevel.Warn, "warning message")
    record = caplog.records.pop()
    assert record.levelname == "WARNING"
    assert "warning message" in record.msg

    # Error message raises
    with pytest.raises(RuntimeError):
        mi.Log(mi.LogLevel.Error, "boom")



"""Test cases with OneDimSolverApp and a Lambertian surface."""

import os

import matplotlib.pyplot as plt
import numpy as np

import eradiate


def test_onedim_lambertian_brf(mode_mono_double, artefact_dir):
    r"""
    Measured lambertian BRF
    =======================

    This test case checks that measured BRF matches the Lambertian surface's
    reflectance.

    Rationale
    ---------

    * Geometry: a square surface with unit size and a Lambertian BRDF with
      reflectance :math:`\rho \in [0.0, 0.3, 0.5, 0.7, 1.0]`.
    * Illumination: Directional illumination with a zenith angle
      :math:`\theta \in [0.0, 30.0, 60.0]°`.
    * Sensor: Multi-distant measure covering a plane (1001 angular points,
      1 sample per pixel) and targeting (0, 0, 0).
      We take an odd number of angular points to ensure that the special value
      :math:`\theta = 0°` is included.


    Expected behaviour
    ------------------

    The BRF results must be equal to reflectance input values.

    Results
    -------

    .. image:: generated/plots/test_onedim_lambertian_brf_0.0.png
       :width: 75%

    .. image:: generated/plots/test_onedim_lambertian_brf_30.0.png
       :width: 75%

    .. image:: generated/plots/test_onedim_lambertian_brf_60.0.png
       :width: 75%
    """
    spp = 1
    n_vza = 51
    illumination_zenith_values = [0.0, 30.0, 60.0]
    reflectance_values = [1.0, 0.7, 0.5, 0.3, 0.0]

    results = {}
    for illumination_zenith in illumination_zenith_values:
        results[illumination_zenith] = {}
        for reflectance in reflectance_values:
            # Run simulation
            exp = eradiate.experiments.AtmosphereExperiment(
                illumination={
                    "type": "directional",
                    "zenith": illumination_zenith,
                    "azimuth": 0.0,
                },
                measures={
                    "type": "distant",
                    "id": "toa_pplane",
                    "construct": "hplane",
                    "zeniths": np.linspace(-90, 90, n_vza),
                    "azimuth": 0.0,
                    "spp": spp,
                },
                surface={
                    "type": "lambertian",
                    "reflectance": reflectance,
                },
                atmosphere=None,
            )
            eradiate.run(exp)

            results[illumination_zenith][reflectance] = exp.results["toa_pplane"]

    # Plot results
    for illumination_zenith in illumination_zenith_values:
        fig = plt.figure(figsize=(6, 3))
        ax1 = plt.gca()

        with plt.rc_context({"lines.linestyle": ":", "lines.marker": "."}):
            for reflectance in reflectance_values:
                results[illumination_zenith][reflectance].brf.plot(ax=ax1, x="vza")

        plt.xlabel("Signed viewing zenith angle [°]")
        plt.xticks([-90.0, -60.0, -30.0, 0.0, 30.0, 60.0, 90.0])
        plt.ylabel("BRF [dimensionless]")
        plt.title(rf"$\theta$ = {illumination_zenith}°")
        plt.legend(
            [f"{reflectance}" for reflectance in reflectance_values],
            title=r"$\rho$",
            loc="center left",
            bbox_to_anchor=(1, 0.5),
        )
        plt.tight_layout()

        outdir = os.path.join(artefact_dir, "plots")
        os.makedirs(outdir, exist_ok=True)
        filename = f"test_onedim_lambertian_brf_{illumination_zenith}.png"
        fname_plot = os.path.join(outdir, filename)
        fig.savefig(fname_plot, dpi=200)

        plt.close()

    for illumination_zenith in illumination_zenith_values:
        for reflectance in reflectance_values:
            assert np.allclose(
                results[illumination_zenith][reflectance].brf, reflectance
            )



import mitsuba as mi
import numpy as np
import pint
import pytest
from pinttr.exceptions import UnitsError

from eradiate import unit_context_kernel as uck
from eradiate import unit_registry as ureg
from eradiate.scenes.core import traverse
from eradiate.scenes.spectra import UniformSpectrum, spectrum_factory
from eradiate.test_tools.types import check_scene_element
from eradiate.units import PhysicalQuantity


@pytest.mark.parametrize(
    "tested, expected",
    [
        (
            {"value": 1.0},
            UniformSpectrum(quantity=None, value=1.0),
        ),
        (
            {"value": 1},
            UniformSpectrum(quantity=None, value=1.0),
        ),
        (
            {
                "value": 1.0 * ureg.m**-1,
                "quantity": PhysicalQuantity.COLLISION_COEFFICIENT,
            },
            UniformSpectrum(
                quantity=PhysicalQuantity.COLLISION_COEFFICIENT,
                value=1.0 * ureg.m**-1,
            ),
        ),
        (
            {"value": 1.0, "quantity": "collision_coefficient"},
            UniformSpectrum(
                quantity=PhysicalQuantity.COLLISION_COEFFICIENT,
                value=1.0 * ureg.m**-1,
            ),
        ),
        (
            {"value": 1.0, "quantity": "speed"},
            KeyError,
        ),
        (
            {"quantity": "collision_coefficient", "value": ureg.Quantity(1.0, "")},
            UnitsError,
        ),
        (
            {"value": 1.0 * ureg.dimensionless},
            ValueError,
        ),
    ],
    ids=[
        "float",
        "int",
        "quantity",
        "quantity_convert",
        "unsupported_quantity",
        "inconsistent_units",
        "unitless_value_but_no_quantity",
    ],
)
def test_uniform_construct_basic(modes_all, tested, expected):
    if isinstance(expected, UniformSpectrum):
        s = UniformSpectrum(**tested)

        assert s.quantity is expected.quantity
        assert np.all(s.value == expected.value)

        if isinstance(s.value, pint.Quantity):
            assert isinstance(s.value.magnitude, float)
        else:
            assert isinstance(s.value, float)

    elif issubclass(expected, Exception):
        with pytest.raises(expected):
            UniformSpectrum(**tested)

    else:
        raise RuntimeError


def test_uniform_kernel_dict(mode_mono):
    # Instantiate from factory using dict
    spectrum = spectrum_factory.convert(
        {
            "type": "uniform",
            "quantity": "radiance",
            "value": 1.0,
            "value_units": "W/m^2/sr/nm",
        }
    )

    # Produced kernel dict is valid
    with uck.override({"radiance": "kW/m^2/sr/nm"}):
        mi_wrapper = check_scene_element(spectrum, mi.Texture)

    # Unit scaling is properly applied
    assert np.allclose(mi_wrapper.parameters["value"], 1e-3)


@pytest.mark.parametrize(
    "quantity, value, w, expected",
    [
        ("dimensionless", 1.0, 550.0, 1.0),
        ("dimensionless", 1.0, [500.0, 600.0] * ureg.nm, 1.0),
        ("collision_coefficient", 1.0, 550.0, 1.0 * ureg.m**-1),
        ("collision_coefficient", 1.0, [500.0, 600.0] * ureg.nm, 1.0 * ureg.m**-1),
    ],
)
def test_uniform_eval_mono(mode_mono, quantity, value, w, expected):
    # No quantity, unitless value
    eval = UniformSpectrum(quantity=quantity, value=value).eval_mono(w)
    assert np.all(expected == eval)
    assert isinstance(eval, pint.Quantity)


def test_integral(mode_mono):
    s = UniformSpectrum(value=0.5)
    # Unitless value is interpreted as dimensionless
    assert s.integral(300.0 * ureg.nm, 400.0 * ureg.nm) == 50.0 * ureg.nm

    s = UniformSpectrum(quantity="collision_coefficient", value=0.5)
    assert s.integral(300.0 * ureg.nm, 400.0 * ureg.nm) == 50.0 * ureg("nm / m")
